from math import floor
from .abstractpathtool import AbstractPathTool
from cadnano2.views import styles

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['Qt', 'QEvent', 'QPointF', 'QRectF'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QColor',
                                       'QFont',
                                       'QFontMetrics',
                                       'QPainterPath',
                                       'QPolygonF',
                                       'QPen'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsLineItem',
                                           'QGraphicsPathItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsSimpleTextItem'])


_baseWidth = styles.PATH_BASE_WIDTH
_pencilcolor = styles.redstroke
_defaultRect = QRectF(0,0, _baseWidth, _baseWidth)
_noPen = QPen(Qt.PenStyle.NoPen)


class PencilTool(AbstractPathTool):
    """
    docstring for PencilTool
    """
    def __init__(self, controller):
        super(PencilTool, self).__init__(controller)
        self._tempXover = ForcedXoverItem(self, None, None)
        self._tempStrandItem = ForcedStrandItem(self, None)
        self._tempStrandItem.hide()
        self._moveIdx = None
        self._isFloatingXoverBegin = True
        self._isDrawingStrand = False

    def __repr__(self):
        return "pencilTool"  # first letter should be lowercase

    def strandItem(self):
        return self._tempStrandItem
    # end def

    def isDrawingStrand(self):
        return self._isDrawingStrand
    # end def

    def setIsDrawingStrand(self, boolval):
        self._isDrawingStrand = boolval
        if boolval == False:
            self._tempStrandItem.hideIt()
    # end def

    def initStrandItemFromVHI(self, virtualHelixItem, strandSet, idx):
        sI = self._tempStrandItem
        self._startIdx = idx
        self._startStrandSet = strandSet
        sI.resetStrandItem(virtualHelixItem, strandSet.isDrawn5to3())
        self._lowDragBound, self._highDragBound = strandSet.getBoundsOfEmptyRegionContaining(idx)
    # end def

    def updateStrandItemFromVHI(self, virtualHelixItem, strandSet, idx):
        sI = self._tempStrandItem
        sIdx = self._startIdx
        if abs(sIdx-idx) > 1 and self.isWithinBounds(idx):
            idxs = (idx, sIdx) if self.isDragLow(idx) else (sIdx, idx)
            sI.strandResizedSlot(idxs)
            sI.showIt()
        # end def
    # end def

    def isDragLow(self, idx):
        sIdx = self._startIdx
        if sIdx-idx > 0:
            return True
        else:
            return False
    # end def

    def isWithinBounds(self, idx):
        return self._lowDragBound <= idx <= self._highDragBound
    # end def

    def attemptToCreateStrand(self, virtualHelixItem, strandSet, idx):
        self._tempStrandItem.hideIt()
        sIdx = self._startIdx
        if abs(sIdx-idx) > 1:
            idx = util.clamp(idx, self._lowDragBound, self._highDragBound)
            idxs = (idx, sIdx) if self.isDragLow(idx) else (sIdx, idx)
            self._startStrandSet.createStrand(*idxs)
    # end def

    def floatingXover(self):
        return self._tempXover
    # end def

    def isFloatingXoverBegin(self):
        return self._isFloatingXoverBegin
    # end def

    def setFloatingXoverBegin(self, boolval):
        self._isFloatingXoverBegin = boolval
        if boolval:
            self._tempXover.hideIt()
        else:
            self._tempXover.showIt()
    # end def

    def attemptToCreateXover(self, virtualHelixItem, strand3p, idx):
        xoi = self._tempXover
        n5 = xoi._node5 
        idx5 = n5._idx
        strand5p = n5._strand
        part = virtualHelixItem.part()
        part.createXover(strand5p, idx5, strand3p, idx)
    # end def
# end class


class ForcedStrandItem(QGraphicsLineItem):
    def __init__(self, tool, virtualHelixItem):
        """The parent should be a VirtualHelixItem."""
        super(ForcedStrandItem, self).__init__(virtualHelixItem)
        self._virtualHelixItem = virtualHelixItem
        self._tool = tool

        isDrawn5to3 = True

        # caps
        self._lowCap = EndpointItem(self, 'low', isDrawn5to3)
        self._highCap = EndpointItem(self, 'high', isDrawn5to3)
        self._lowCap.disableEvents()
        self._highCap.disableEvents()
        
        # orientation
        self._isDrawn5to3 = isDrawn5to3

        # create a larger click area rect to capture mouse events
        self._clickArea = cA = QGraphicsRectItem(_defaultRect, self)
        cA.mousePressEvent = self.mousePressEvent
        cA.setPen(_noPen)

        # cA.setBrush(QBrush(Qt.white))
        self.setZValue(styles.ZENDPOINTITEM+1)
        cA.setZValue(styles.ZENDPOINTITEM)
        self._lowCap.setZValue(styles.ZENDPOINTITEM+2)
        self._highCap.setZValue(styles.ZENDPOINTITEM+2)
        cA.setFlag(QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent)

        self._updatePensAndBrushes()
        self.hideIt()
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def strandResizedSlot(self, idxs):
        """docstring for strandResizedSlot"""
        lowMoved = self._lowCap.updatePosIfNecessary(idxs[0])
        highMoved = self._highCap.updatePosIfNecessary(idxs[1])
        if lowMoved:
            self.updateLine(self._lowCap)
        if highMoved:
            self.updateLine(self._highCap)
    # end def

    def strandRemovedSlot(self):
        scene = self.scene()
        scene.removeItem(self._clickArea)
        scene.removeItem(self._highCap)
        scene.removeItem(self._lowCap)

        self._clickArea = None
        self._highCap = None
        self._lowCap = None

        scene.removeItem(self)
    # end def

    ### ACCESSORS ###

    def virtualHelixItem(self):
        return self._virtualHelixItem

    def activeTool(self):
        return self._tool
    # end def

    def hideIt(self):
        self.hide()
        self._lowCap.hide()
        self._highCap.hide()
        self._clickArea.hide()
    # end def

    def showIt(self):
        self._lowCap.show()
        self._highCap.show()
        self._clickArea.show()
        self.show()
    # end def

    def resetStrandItem(self, virtualHelixItem, isDrawn5to3):
        self.setParentItem(virtualHelixItem)
        self._virtualHelixItem = virtualHelixItem
        self.resetEndPointItems(isDrawn5to3)
    # end def

    def resetEndPointItems(self, isDrawn5to3):
        bw = _baseWidth
        self._isDrawn5to3 = isDrawn5to3
        self._lowCap.resetEndPoint(isDrawn5to3)
        self._highCap.resetEndPoint(isDrawn5to3)
        line = self.line()
        p1 = line.p1()
        p2 = line.p2()
        if isDrawn5to3:
            p1.setY(bw/2)
            p2.setY(bw/2)
            self._clickArea.setY(0)
        else:
            p1.setY(3*bw/2)
            p2.setY(3*bw/2)
            self._clickArea.setY(bw)
        line.setP1(p1)
        line.setP2(p2)
        self.setLine(line)
    # end def

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def updateLine(self, movedCap):
        # setup
        bw = _baseWidth
        cA = self._clickArea
        line = self.line()
        # set new line coords
        if movedCap == self._lowCap:
            p1 = line.p1()
            newX = self._lowCap.pos().x() + bw
            p1.setX(newX)
            line.setP1(p1)
            temp = cA.rect()
            temp.setLeft(newX-bw)
            cA.setRect(temp)
        else:
            p2 = line.p2()
            newX = self._highCap.pos().x()
            p2.setX(newX)
            line.setP2(p2)
            temp = cA.rect()
            temp.setRight(newX+bw)
            cA.setRect(temp)
        self.setLine(line)
    # end def

    def _updatePensAndBrushes(self):
        color = QColor(_pencilcolor)
        penWidth = styles.PATH_STRAND_STROKE_WIDTH
        pen = QPen(color, penWidth)
        brush = QBrush(color)
        pen.setCapStyle(Qt.PenCapStyle.FlatCap)
        self.setPen(pen)
        self._lowCap.setBrush(brush)
        self._highCap.setBrush(brush)
    # end def
# end class

_toHelixNumFont = styles.XOVER_LABEL_FONT
# precalculate the height of a number font.  Assumes a fixed font
# and that only numbers will be used for labels
_fm = QFontMetrics(_toHelixNumFont)
_enabbrush = QBrush(Qt.BrushStyle.SolidPattern)  # Also for the helix number label
_noBrush = QBrush(Qt.BrushStyle.NoBrush)

# _rect = QRectF(0, 0, baseWidth, baseWidth)
_xScale = styles.PATH_XOVER_LINE_SCALE_X  # control point x constant
_yScale = styles.PATH_XOVER_LINE_SCALE_Y  # control point y constant
_rect = QRectF(0, 0, _baseWidth, _baseWidth)
_blankRect = QRectF(0, 0, 2*_baseWidth, _baseWidth)

ppL5 = QPainterPath()  # Left 5' PainterPath
ppR5 = QPainterPath()  # Right 5' PainterPath
ppL3 = QPainterPath()  # Left 3' PainterPath
ppR3 = QPainterPath()  # Right 3' PainterPath

# set up ppL5 (left 5' blue square)
ppL5.addRect(0.25*_baseWidth, 0.125*_baseWidth,0.75*_baseWidth, 0.75*_baseWidth)
# set up ppR5 (right 5' blue square)
ppR5.addRect(0, 0.125*_baseWidth, 0.75*_baseWidth, 0.75*_baseWidth)
# set up ppL3 (left 3' blue triangle)
l3poly = QPolygonF()
l3poly.append(QPointF(_baseWidth, 0))
l3poly.append(QPointF(0.25*_baseWidth, 0.5*_baseWidth))
l3poly.append(QPointF(_baseWidth, _baseWidth))
ppL3.addPolygon(l3poly)
# set up ppR3 (right 3' blue triangle)
r3poly = QPolygonF()
r3poly.append(QPointF(0, 0))
r3poly.append(QPointF(0.75*_baseWidth, 0.5*_baseWidth))
r3poly.append(QPointF(0, _baseWidth))
ppR3.addPolygon(r3poly)

class ForcedXoverNode3(QGraphicsRectItem):
    """
    This is a QGraphicsRectItem to allow actions and also a 
    QGraphicsSimpleTextItem to allow a label to be drawn
    """
    def __init__(self, virtualHelixItem, xoverItem, strand3p, idx):
        super(ForcedXoverNode3, self).__init__(virtualHelixItem)
        self._vhi = virtualHelixItem
        self._xoverItem = xoverItem
        self._idx = idx
        self._isOnTop = virtualHelixItem.isStrandOnTop(strand3p)
        self._isDrawn5to3 = strand3p.strandSet().isDrawn5to3()
        self._strandType = strand3p.strandSet().strandType()

        self._partnerVirtualHelix = virtualHelixItem

        self._blankThing = QGraphicsRectItem(_blankRect, self)
        self._blankThing.setBrush(QBrush(Qt.GlobalColor.white))
        self._pathThing = QGraphicsPathItem(self)
        self.configurePath()

        self.setPen(_noPen)
        self._label = None
        self.setPen(_noPen)
        self.setBrush(_noBrush)
        self.setRect(_rect)

        self.setZValue(styles.ZENDPOINTITEM+1)
    # end def

    def updateForFloatFromVHI(self, virtualHelixItem, strandType, idxX, idxY):
        """

        """
        self._vhi = virtualHelixItem
        self.setParentItem(virtualHelixItem)
        self._strandType = strandType
        self._idx = idxX
        self._isOnTop = self._isDrawn5to3 = True if idxY == 0 else False
        self.updatePositionAndAppearance(isFromStrand=False)
    # end def

    def updateForFloatFromStrand(self, virtualHelixItem, strand3p, idx):
        """

        """
        self._vhi = virtualHelixItem
        self._strand = strand3p
        self.setParentItem(virtualHelixItem)
        self._idx = idx
        self._isOnTop = virtualHelixItem.isStrandOnTop(strand3p)
        self._isDrawn5to3 = strand3p.strandSet().isDrawn5to3()
        self._strandType = strand3p.strandSet().strandType()
        self.updatePositionAndAppearance()
    # end def

    def strandType(self):
        return self._strandType
    # end def

    def configurePath(self):
        self._pathThing.setBrush(QBrush(styles.redstroke))
        path = ppR3 if self._isDrawn5to3 else ppL3
        offset = -_baseWidth if self._isDrawn5to3 else _baseWidth
        self._pathThing.setPath(path)
        self._pathThing.setPos(offset, 0)

        offset = -_baseWidth if self._isDrawn5to3 else 0
        self._blankThing.setPos(offset, 0)

        self._blankThing.show()
        self._pathThing.show()
    # end def

    def refreshXover(self):
        self._xoverItem.refreshXover()
    # end def

    def setPartnerVirtualHelix(self, virtualHelixItem):
        self._partnerVirtualHelix = virtualHelixItem
    # end def

    def idx(self):
        return self._idx
    # end def

    def virtualHelixItem(self):
        return self._vhi
    # end def

    def point(self):
        return self._vhi.upperLeftCornerOfBaseType(self._idx, self._strandType)
    # end def

    def floatPoint(self):
        pt = self.pos()
        return pt.x(), pt.y()
    # end def

    def isOnTop(self):
        return self._isOnTop
    # end def

    def isDrawn5to3(self):
        return self._isDrawn5to3
    # end def

    def updatePositionAndAppearance(self, isFromStrand=True):
        """
        Sets position by asking the VirtualHelixItem
        Sets appearance by choosing among pre-defined painterpaths (from
        normalstrandgraphicsitem) depending on drawing direction.
        """
        self.setPos(*self.point())
        n5 = self._xoverItem._node5
        if isFromStrand:
            fromStrand, fromIdx = (n5._strand, n5._idx) if n5 != self else (None, None)
            if self._strand.canInstallXoverAt(self._idx, fromStrand, fromIdx):
                self.configurePath()
                # We can only expose a 5' end. But on which side?
                isLeft = True if self._isDrawn5to3 else False
                self._updateLabel(isLeft)
            else:
                self.hideItems()
        else:
            self.hideItems()
    # end def

    def updateConnectivity(self):
        isLeft = True if self._isDrawn5to3 else False
        self._updateLabel(isLeft)
    # end def

    def remove(self):
        """
        Clean up this joint
        """
        scene = self.scene()
        scene.removeItem(self._label)
        self._label = None
        scene.removeItem(self._pathThing)
        self._pathThing = None
        scene.removeItem(self._blankThing)
        self._blankThing = None
        scene.removeItem(self)
    # end def

    def _updateLabel(self, isLeft):
        """
        Called by updatePositionAndAppearance during init, or later by
        updateConnectivity. Updates drawing and position of the label.
        """
        lbl = self._label
        if self._idx != None:
            bw = _baseWidth
            num = self._partnerVirtualHelix.number()
            tBR = _fm.tightBoundingRect(str(num))
            halfLabelH = tBR.height()/2.0
            halfLabelW = tBR.width()/2.0
            # determine x and y positions
            labelX = bw/2.0 - halfLabelW
            if self._isOnTop:
                labelY = -0.25*halfLabelH - 0.5 - 0.5*bw
            else:
                labelY = 2*halfLabelH + 0.5 + 0.5*bw
            # adjust x for left vs right
            labelXoffset = 0.25*bw if isLeft else -0.25*bw
            labelX += labelXoffset
            # adjust x for numeral 1
            if num == 1: labelX -= halfLabelW/2.0
            # create text item
            if lbl == None:
                lbl = QGraphicsSimpleTextItem(str(num), self)
            lbl.setPos(labelX, labelY)
            lbl.setBrush(_enabbrush)
            lbl.setFont(_toHelixNumFont)
            self._label = lbl

            lbl.setText( str(self._partnerVirtualHelix.number()) )
            lbl.show()
        # end if
    # end def

    def hideItems(self):
        if self._label:
            self._label.hide()
        if self._blankThing:
            self._pathThing.hide()
        if self._blankThing:
            self._blankThing.hide()
    # end def
# end class


class ForcedXoverNode5(ForcedXoverNode3):
    """
    XoverNode5 is the partner of XoverNode3. It dif
    XoverNode3 handles:
    1. Drawing of the 5' end of an xover, and its text label. Drawing style
    is determined by the location of the xover with in a vhelix (is it a top
    or bottom vstrand?).
    2. Notifying XoverStrands in the model when connectivity changes.

    """
    def __init__(self, virtualHelixItem, xoverItem, strand5p, idx):
        super(ForcedXoverNode5, self).__init__(virtualHelixItem, xoverItem, strand5p, idx)
    # end def

    def configurePath(self):
        self._pathThing.setBrush(QBrush(styles.redstroke))
        path = ppL5 if self._isDrawn5to3 else ppR5
        offset = _baseWidth if self._isDrawn5to3 else -_baseWidth
        self._pathThing.setPath(path)
        self._pathThing.setPos(offset, 0)

        offset = 0 if self._isDrawn5to3 else -_baseWidth
        self._blankThing.setPos(offset, 0)

        self._blankThing.show()
        self._pathThing.show()
    # end def

    def updatePositionAndAppearance(self, isFromStrand=True):
        """Same as XoverItem3, but exposes 3' end"""
        self.setPos(*self.point())
        self.configurePath()
        # # We can only expose a 3' end. But on which side?
        isLeft = False if self._isDrawn5to3 else True
        self._updateLabel(isLeft)
    # end def
# end class

class ForcedXoverItem(QGraphicsPathItem):
    """
    This class handles:
    1. Drawing the spline between the XoverNode3 and XoverNode5 graphics
    items in the path view.

    XoverItem should be a child of a PartItem.
    """

    def __init__(self, tool, partItem, virtualHelixItem):
        """
        strandItem is a the model representation of the 5prime most strand
        of a Xover
        """
        super(ForcedXoverItem, self).__init__(partItem)
        self._tool = tool
        self._virtualHelixItem = virtualHelixItem
        self._strandType = None
        self._node5 = None
        self._node3 = None
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable) # for keyPressEvents
        self.setZValue(styles.ZPATHTOOL)
        self.hide()
    # end def

    ### SLOTS ###

    ### METHODS ###
    def remove(self):
        scene = self.scene()
        if self._node3:
            scene.removeItem(self._node3)
            scene.removeItem(self._node5)
        scene.removeItem(self)
    # end def

    def strandType(self):
        return self._strandType
    # end def
    
    def hide5prime(self):
        self._node5._pathThing.hide()

    def hide3prime(self):
        self._node3._pathThing.hide()
        
    def show3prime(self):
        if self._node3._blankThing.isVisible():
            self._node3._pathThing.show()

    def hideIt(self):
        self.hide()
        self.clearFocus()
        if self._node3:
            self._node3.hide()
            self._node5.hide()
    # end def

    def showIt(self):
        self.show()
        self.setFocus()
        if self._node3:
            self._node3.show()
            self._node5.show()
    # end def

    def keyPressEvent(self, event):
        """
        Must intercept invalid input events.  Make changes here
        """
        a = event.key()
        if a in [Qt.Key.Key_Control, Qt.Key.Key_Left, Qt.Key.Key_Right, Qt.Key.Key_Up, Qt.Key.Key_Down]:
            QGraphicsPathItem.keyPressEvent(self, event)
        else:
            self._tool.setFloatingXoverBegin(True)
    # end def

    def updateBase(self, virtualHelixItem, strand5p, idx):
        # floating Xover!
        self._virtualHelixItem = virtualHelixItem
        self.setParentItem(virtualHelixItem.partItem())
        self._strandType = strand5p.strandSet().strandType()
        if self._node5 == None:
            self._node5 = ForcedXoverNode5(virtualHelixItem, self, strand5p, idx)
            self._node3 = ForcedXoverNode3(virtualHelixItem, self, strand5p, idx)
        self._node5.updateForFloatFromStrand(virtualHelixItem, strand5p, idx)
        self._node3.updateForFloatFromStrand(virtualHelixItem, strand5p, idx)
        self.updateFloatPath()
    # end def

    def updateFloatingFromVHI(self, virtualHelixItem, strandType, idxX, idxY):
        # floating Xover!
        self._node5.setPartnerVirtualHelix(virtualHelixItem)
        self._node5.updatePositionAndAppearance()
        self._node3.setPartnerVirtualHelix(self._virtualHelixItem)
        self._node3.updateForFloatFromVHI(virtualHelixItem, strandType, idxX, idxY)
        self.updateFloatPath()
    # end def

    def updateFloatingFromStrandItem(self, virtualHelixItem, strand3p, idx):
        # floating Xover!
        self._node3.updateForFloatFromStrand(virtualHelixItem, strand3p, idx)
        self.updateFloatPath()
    # end def

    def updateFloatingFromPartItem(self, partItem, pt):
        self._node3.hideItems()
        self.updateFloatPath(pt)
    # end def

    def updateFloatPath(self, point=None):
        """
        Draws a quad curve from the edge of the fromBase
        to the top or bottom of the toBase (q5), and
        finally to the center of the toBase (toBaseEndpoint).

        If floatPos!=None, this is a floatingXover and floatPos is the
        destination point (where the mouse is) while toHelix, toIndex
        are potentially None and represent the base at floatPos.

        """
        node3 = self._node3
        node5 = self._node5

        bw = _baseWidth

        vhi5 = self._virtualHelixItem
        partItem = vhi5.partItem()
        pt5 = vhi5.mapToItem(partItem, *node5.floatPoint())

        fiveIsTop = node5.isOnTop()
        fiveIs5to3 = node5.isDrawn5to3()

        # Enter/exit are relative to the direction that the path travels
        # overall.
        fiveEnterPt = pt5 + QPointF(0 if fiveIs5to3 else 1, .5)*bw
        fiveCenterPt = pt5 + QPointF(.5, .5)*bw
        fiveExitPt = pt5 + QPointF(.5, 0 if fiveIsTop else 1)*bw

        vhi3 = node3.virtualHelixItem()

        if point:
            pt3 = point
            threeIsTop = True
            threeIs5to3 = True
            sameStrand = False
            sameParity = False
            threeEnterPt = threeCenterPt = threeExitPt = pt3
        else: 
            pt3 = vhi3.mapToItem(partItem, *node3.point())
            threeIsTop = node3.isOnTop()
            threeIs5to3 = node3.isDrawn5to3()
            sameStrand = (node5.strandType() == node3.strandType()) and vhi3 == vhi5
            sameParity = fiveIs5to3 == threeIs5to3

            threeEnterPt = pt3 + QPointF(.5, 0 if threeIsTop else 1)*bw
            threeCenterPt = pt3 + QPointF(.5, .5)*bw
            threeExitPt = pt3 + QPointF(1 if threeIs5to3 else 0, .5)*bw

        c1 = QPointF()
        # case 1: same strand
        if sameStrand:
            dx = abs(threeEnterPt.x() - fiveExitPt.x())
            c1.setX(0.5 * (fiveExitPt.x() + threeEnterPt.x()))
            if fiveIsTop:
                c1.setY(fiveExitPt.y() - _yScale * dx)
            else:
                c1.setY(fiveExitPt.y() + _yScale * dx)
            # case 2: same parity
        elif sameParity:
             dy = abs(threeEnterPt.y() - fiveExitPt.y())
             c1.setX(fiveExitPt.x() + _xScale * dy)
             c1.setY(0.5 * (fiveExitPt.y() + threeEnterPt.y()))
        # case 3: different parity
        else:
            if fiveIsTop and fiveIs5to3:
                c1.setX(fiveExitPt.x() - _xScale *\
                        abs(threeEnterPt.y() - fiveExitPt.y()))
            else:
                c1.setX(fiveExitPt.x() + _xScale *\
                        abs(threeEnterPt.y() - fiveExitPt.y()))
            c1.setY(0.5 * (fiveExitPt.y() + threeEnterPt.y()))

        # Construct painter path
        painterpath = QPainterPath()
        painterpath.moveTo(fiveEnterPt)
        painterpath.lineTo(fiveCenterPt)
        painterpath.lineTo(fiveExitPt)
        painterpath.quadTo(c1, threeEnterPt)
        painterpath.lineTo(threeCenterPt)
        painterpath.lineTo(threeExitPt)

        self.setPath(painterpath)
        self._updateFloatPen()
    # end def

    # def _updatePen(self, strand5p):
    #     oligo = strand5p.oligo()
    #     color = QColor(oligo.color())
    #     penWidth = styles.PATH_STRAND_STROKE_WIDTH
    #     if oligo.shouldHighlight():
    #         penWidth = styles.PATH_STRAND_HIGHLIGHT_STROKE_WIDTH
    #         color.setAlpha(128)
    #     pen = QPen(color, penWidth)
    #     pen.setCapStyle(Qt.PenCapStyle.FlatCap)
    #     self.setPen(pen)
    # # end def

    def _updateFloatPen(self):
        penWidth = styles.PATH_STRAND_STROKE_WIDTH
        pen = QPen(_pencilcolor, penWidth)
        pen.setCapStyle(Qt.PenCapStyle.FlatCap)
        self.setPen(pen)
    # end def
# end class XoverItem

util.qtWrapImport('QtGui', globals(), ['QPolygonF'])

ppL5 = QPainterPath()  # Left 5' PainterPath
ppR5 = QPainterPath()  # Right 5' PainterPath
ppL3 = QPainterPath()  # Left 3' PainterPath
ppR3 = QPainterPath()  # Right 3' PainterPath
pp53 = QPainterPath()  # Left 5', Right 3' PainterPath
pp35 = QPainterPath()  # Left 5', Right 3' PainterPath
# set up ppL5 (left 5' blue square)
ppL5.addRect(0.25*_baseWidth, 0.125*_baseWidth,0.75*_baseWidth, 0.75*_baseWidth)
# set up ppR5 (right 5' blue square)
ppR5.addRect(0, 0.125*_baseWidth, 0.75*_baseWidth, 0.75*_baseWidth)
# set up ppL3 (left 3' blue triangle)
l3poly = QPolygonF()
l3poly.append(QPointF(_baseWidth, 0))
l3poly.append(QPointF(0.25*_baseWidth, 0.5*_baseWidth))
l3poly.append(QPointF(_baseWidth, _baseWidth))
ppL3.addPolygon(l3poly)
# set up ppR3 (right 3' blue triangle)
r3poly = QPolygonF()
r3poly.append(QPointF(0, 0))
r3poly.append(QPointF(0.75*_baseWidth, 0.5*_baseWidth))
r3poly.append(QPointF(0, _baseWidth))
ppR3.addPolygon(r3poly)

# single base left 5'->3'
pp53.addRect(0, 0.125*_baseWidth, 0.5*_baseWidth, 0.75*_baseWidth)
poly53 = QPolygonF()
poly53.append(QPointF(0.5*_baseWidth, 0))
poly53.append(QPointF(_baseWidth, 0.5*_baseWidth))
poly53.append(QPointF(0.5*_baseWidth, _baseWidth))
pp53.addPolygon(poly53)
# single base left 3'<-5'
pp35.addRect(0.50*_baseWidth, 0.125*_baseWidth, 0.5*_baseWidth, 0.75*_baseWidth)
poly35 = QPolygonF()
poly35.append(QPointF(0.5*_baseWidth, 0))
poly35.append(QPointF(0, 0.5*_baseWidth))
poly35.append(QPointF(0.5*_baseWidth, _baseWidth))
pp35.addPolygon(poly35)

class EndpointItem(QGraphicsPathItem):
    def __init__(self, strandItem, captype, isDrawn5to3):
        """The parent should be a StrandItem."""
        super(EndpointItem, self).__init__(strandItem.virtualHelixItem())

        self._strandItem = strandItem
        self._activeTool = strandItem.activeTool()
        self._capType = captype
        self._lowDragBound = None
        self._highDragBound = None
        self._initCapSpecificState(isDrawn5to3)
        self.setPen(_noPen)
        # for easier mouseclick
        self._clickArea = cA = QGraphicsRectItem(_defaultRect, self)
        self._clickArea.setAcceptHoverEvents(True)
        cA.hoverMoveEvent = self.hoverMoveEvent
        cA.mousePressEvent = self.mousePressEvent
        cA.mouseMoveEvent = self.mouseMoveEvent
        cA.setPen(_noPen)
        
    # end def

    def __repr__(self):
        return "%s" % self.__class__.__name__

    ### SIGNALS ###

    ### SLOTS ###

    ### ACCESSORS ###
    def idx(self):
        """Look up baseIdx, as determined by strandItem idxs and cap type."""
        if self._capType == 'low':
            return self._strandItem.idxs()[0]
        else:  # high or dual, doesn't matter
            return self._strandItem.idxs()[1]
    # end def
    
    def partItem(self):
        return self._strandItem.partItem()
    # end def

    def disableEvents(self):
        self._clickArea.setAcceptHoverEvents(False)
        self.mouseMoveEvent = QGraphicsPathItem.mouseMoveEvent
        self.mousePressEvent = QGraphicsPathItem.mousePressEvent
    # end def

    def window(self):
        return self._strandItem.window()

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def updatePosIfNecessary(self, idx):
        """Update position if necessary and return True if updated."""
        x = int(idx*_baseWidth)
        if x != self.x():
            self.setPos(x, self.y())
            return True
        return False

    def resetEndPoint(self, isDrawn5to3):
        self.setParentItem(self._strandItem.virtualHelixItem())
        self._initCapSpecificState(isDrawn5to3)
        upperLeftY = 0 if isDrawn5to3 else _baseWidth
        self.setY(upperLeftY)
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _initCapSpecificState(self, isDrawn5to3):
        cT = self._capType
        if cT == 'low':
            path = ppL5 if isDrawn5to3 else ppL3
        elif cT == 'high':
            path = ppR3 if isDrawn5to3 else ppR5
        elif cT == 'dual':
            path = pp53 if isDrawn5to3 else pp35
        self.setPath(path)
    # end def

    def _getNewIdxsForResize(self, baseIdx):
        """Returns a tuple containing idxs to be passed to the """
        cT = self._capType
        if cT == 'low':
            return (baseIdx, self._strandItem.idxs()[1])
        elif cT == 'high':
            return (self._strandItem.idxs()[0], baseIdx)
        elif cT == 'dual':
            raise NotImplementedError

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent, calling the approproate tool method as
        necessary. Stores _moveIdx for future comparison.
        """
        self.scene().views()[0].addToPressList(self)
        self._strandItem.virtualHelixItem().setActive(self.idx())
        self._moveIdx = self.idx()
        activeToolStr = str(self._activeTool())
        if activeToolStr == 'pencilTool':
            return self._strandItem.pencilToolMousePress(self.idx())
        toolMethodName = activeToolStr + "MousePress"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            getattr(self, toolMethodName)(modifiers)  # call tool method

    def hoverMoveEvent(self, event):
        """
        Parses a mousePressEvent, calling the approproate tool method as
        necessary. Stores _moveIdx for future comparison.
        """
        activeToolStr = str(self._activeTool())
        if activeToolStr == 'pencilTool':
            return self._strandItem.pencilToolHoverMove(self.idx())

    def mouseMoveEvent(self, event):
        """
        Parses a mouseMoveEvent, calling the approproate tool method as
        necessary. Updates _moveIdx if it changed.
        """
        toolMethodName = str(self._activeTool()) + "MouseMove"
        if hasattr(self, toolMethodName):  # if the tool method exists
            idx = int(floor((self.x()+event.position().x()) / _baseWidth))
            if idx != self._moveIdx:  # did we actually move?
                modifiers = event.modifiers()
                self._moveIdx = idx
                getattr(self, toolMethodName)(modifiers, idx)  # call tool method

    def customMouseRelease(self, event):
        """
        Parses a mouseReleaseEvent, calling the approproate tool method as
        necessary. Deletes _moveIdx if necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseRelease"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            x = event.position().x()
            getattr(self, toolMethodName)(modifiers, x)  # call tool method
        if self._moveIdx is not None:
            self._moveIdx = None
# end class


from cadnano2.controllers.itemcontrollers.strand.abstractstranditemcontroller \
     import AbstractStrandItemController


class EndpointItemController(AbstractStrandItemController):
    def __init__(self, strandItem, modelOligo, modelStrand):
        super(EndpointItemController, self).__init__(strandItem, modelStrand)
        self.connectSignals()
    # end def



#!/usr/bin/env python

"""
main.py

Created by Shawn Douglas on 2010-09-26.
"""

import sys
import os
import pkg_resources  # part of setuptools
sys.path.insert(0, '.')
from . import cadnano

if "-t" in sys.argv:
    os.environ['CADNANO_IGNORE_ENV_VARS_EXCEPT_FOR_ME'] = 'YES'

cadnano.initAppWithGui()

welcome_message = """
Thank you for using Cadnano2 ({})!

We invite you to support the project by citing this reference:

Rapid prototyping of 3D DNA-origami shapes with caDNAno
Douglas et al. Nucleic Acids Res: 37(15):5001–6 (2009)
https://doi.org/10.1093/nar/gkp436

Report bugs at https://github.com/douglaslab/cadnano2 (include terminal output)
Contact: shawn.douglas [at] ucsf.edu
"""

version = pkg_resources.require("cadnano2")[0].version

def main(args=None):
    print(welcome_message.format(version))
    app = cadnano.app()
    if "-p" in sys.argv:
        print("Collecting profile data into cadnano.profile")
        import cProfile
        cProfile.run('app.exec_()', 'cadnano.profile')
        print("Done collecting profile data. Use -P to print it out.")
        exit()
    elif "-P" in sys.argv:
        from pstats import Stats
        s = Stats('cadnano.profile')
        print("Internal Time Top 10:")
        s.sort_stats('cumulative').print_stats(10)
        print("")
        print("Total Time Top 10:")
        s.sort_stats('time').print_stats(10)
        exit()
    elif "-t" in sys.argv:
        print("running tests")
        from tests.runall import main as runTests
        runTests(useXMLRunner=False)
        exit()
    app.exec_()


if __name__ == '__main__':
    sys.exit(main() or 0)


class PartItemController(object):
    def __init__(self, partItem, modelPart):
        self._partItem = partItem
        self._modelPart = modelPart
        self.connectSignals()

    def connectSignals(self):
        mP = self._modelPart
        pI = self._partItem

        if hasattr(pI, "partHideSlot"):
            mP.partHideSignal.connect(pI.partHideSlot)
        if hasattr(pI, "partActiveVirtualHelixChangedSlot"):
            mP.partActiveVirtualHelixChangedSignal.connect(pI.partActiveVirtualHelixChangedSlot)

        mP.partDimensionsChangedSignal.connect(pI.partDimensionsChangedSlot)
        mP.partParentChangedSignal.connect(pI.partParentChangedSlot)
        mP.partPreDecoratorSelectedSignal.connect(pI.partPreDecoratorSelectedSlot)
        mP.partRemovedSignal.connect(pI.partRemovedSlot)
        mP.partStrandChangedSignal.connect(pI.updatePreXoverItemsSlot)
        mP.partVirtualHelixAddedSignal.connect(pI.partVirtualHelixAddedSlot)
        mP.partVirtualHelixRenumberedSignal.connect(pI.partVirtualHelixRenumberedSlot)
        mP.partVirtualHelixResizedSignal.connect(pI.partVirtualHelixResizedSlot)
        mP.partVirtualHelicesReorderedSignal.connect(pI.partVirtualHelicesReorderedSlot)
    # end def

    def disconnectSignals(self):
        mP = self._modelPart
        pI = self._partItem

        if hasattr(pI, "partHideSlot"):
            mP.partHideSignal.disconnect(pI.partHideSlot)
        if hasattr(pI, "partActiveVirtualHelixChangedSlot"):
            mP.partActiveVirtualHelixChangedSignal.disconnect(pI.partActiveVirtualHelixChangedSlot)

        mP.partDimensionsChangedSignal.disconnect(pI.partDimensionsChangedSlot)
        mP.partParentChangedSignal.disconnect(pI.partParentChangedSlot)
        mP.partPreDecoratorSelectedSignal.disconnect(pI.partPreDecoratorSelectedSlot)
        mP.partRemovedSignal.disconnect(pI.partRemovedSlot)
        mP.partStrandChangedSignal.disconnect(pI.updatePreXoverItemsSlot)
        mP.partVirtualHelixAddedSignal.disconnect(pI.partVirtualHelixAddedSlot)
        mP.partVirtualHelixRenumberedSignal.disconnect(pI.partVirtualHelixRenumberedSlot)
        mP.partVirtualHelixResizedSignal.disconnect(pI.partVirtualHelixResizedSlot)
        mP.partVirtualHelicesReorderedSignal.disconnect(pI.partVirtualHelicesReorderedSlot)
    # end def
# end class



# Form implementation generated from reading ui file 'dialogs/warning.ui'
#
# Created by: PyQt6 UI code generator 6.1.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Warning(object):
    def setupUi(self, Warning):
        Warning.setObjectName("Warning")
        Warning.setWindowModality(QtCore.Qt.WindowModality.WindowModality.ApplicationModal)
        Warning.resize(400, 300)
        self.buttonBox = QtWidgets.QDialogButtonBox(Warning)
        self.buttonBox.setGeometry(QtCore.QRect(30, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.title = QtWidgets.QLabel(Warning)
        self.title.setGeometry(QtCore.QRect(20, 30, 361, 31))
        font = QtGui.QFont()
        font.setPointSize(24)
        self.title.setFont(font)
        self.title.setObjectName("title")
        self.message = QtWidgets.QLabel(Warning)
        self.message.setGeometry(QtCore.QRect(20, 80, 361, 151))
        self.message.setWordWrap(True)
        self.message.setObjectName("message")

        self.retranslateUi(Warning)
        self.buttonBox.accepted.connect(Warning.accept)
        self.buttonBox.rejected.connect(Warning.reject)
        QtCore.QMetaObject.connectSlotsByName(Warning)

    def retranslateUi(self, Warning):
        _translate = QtCore.QCoreApplication.translate
        Warning.setWindowTitle(_translate("Warning", "Dialog"))
        self.title.setText(_translate("Warning", "Warning"))
        self.message.setText(_translate("Warning", "Text here."))



import cadnano2.util as util


class XoverItemController(object):
    def __init__(self, xoverItem, modelStrand5p):
        self._xoverItem = xoverItem
        self._modelStrand5p = modelStrand5p
        self._modelOligo = modelStrand5p.oligo()
        self.connectSignals()
    # end def

    def reconnectOligoSignals(self):
        """
        use this for whenever a strands oligo changes
        """
        self.disconnectSignals()
        self.connectSignals()
    # end def

    def reconnectSignals(self, strand):
        self.disconnectSignals()
        self._modelStrand5p = strand
        self.connectSignals()
    # end def

    def connectSignals(self):
        xI = self._xoverItem
        s5p = self._modelStrand5p
        mO = s5p.oligo()
        self._modelOligo = mO

        s5p.strand5pHasSwappedSignal.connect(xI.strandSwapSlot)
        s5p.strandHasNewOligoSignal.connect(xI.strandHasNewOligoSlot)
        mO.oligoAppearanceChangedSignal.connect(xI.oligoAppearanceChangedSlot)
        s5p.strandXover5pRemovedSignal.connect(xI.xover5pRemovedSlot)
    # end def

    def disconnectSignals(self):
        xI = self._xoverItem
        s5p = self._modelStrand5p
        mO = self._modelOligo

        s5p.strand5pHasSwappedSignal.disconnect(xI.strandSwapSlot)
        s5p.strandHasNewOligoSignal.disconnect(xI.strandHasNewOligoSlot)
        mO.oligoAppearanceChangedSignal.disconnect(xI.oligoAppearanceChangedSlot)
        s5p.strandXover5pRemovedSignal.connect(xI.xover5pRemovedSlot)
    # end def



import cadnano2.util as util
from cadnano2.controllers.itemcontrollers.strand.abstractstranditemcontroller \
     import AbstractStrandItemController


class StrandItemController(AbstractStrandItemController):
    def __init__(self, strandItem, modelStrand):
        super(StrandItemController, self).__init__(strandItem, modelStrand)
        self.connectSignals()
    # end def

    def reconnectOligoSignals(self):
        """
        use this for whenever a strands oligo changes
        """
        AbstractStrandItemController.disconnectOligoSignals(self)
        self.disconnectOligoSignals()
        AbstractStrandItemController.connectOligoSignals(self)
        self.connectOligoSignals()
    # end def

    def connectSignals(self):
        AbstractStrandItemController.connectSignals(self)
        mS = self._modelStrand
        sI = self._strandItem
        mS.strandResizedSignal.connect(sI.strandResizedSlot)
        # mS.strandXover5pChangedSignal.connect(sI.strandXover5pChangedSlot)
        mS.strandUpdateSignal.connect(sI.strandUpdateSlot)
        self.connectOligoSignals()
    # end def

    def connectOligoSignals(self):
        mO = self._modelStrand.oligo()
        self._modelOligo = mO
        sI = self._strandItem
        mO.oligoSequenceAddedSignal.connect(sI.oligoSequenceAddedSlot)
        mO.oligoSequenceClearedSignal.connect(sI.oligoSequenceClearedSlot)
    # end def

    def disconnectSignals(self):
        AbstractStrandItemController.disconnectSignals(self)
        mS = self._modelStrand
        sI = self._strandItem
        mS.strandResizedSignal.disconnect(sI.strandResizedSlot)
        # mS.strandXover5pChangedSignal.disconnect(sI.strandXover5pChangedSlot)
        mS.strandUpdateSignal.disconnect(sI.strandUpdateSlot)
        self.disconnectOligoSignals()
    # end def

    def disconnectOligoSignals(self):
        mO = self._modelOligo
        sI = self._strandItem
        mO.oligoSequenceAddedSignal.disconnect(sI.oligoSequenceAddedSlot)
        mO.oligoSequenceClearedSignal.disconnect(sI.oligoSequenceClearedSlot)
    # end def



#!/usr/bin/env python
# encoding: utf-8

# from exceptions import NotImplementedError
import cadnano2.util as util
util.qtWrapImport('QtGui', globals(), ['QGraphicsPathItem'])


class AbstractStrandItem(QGraphicsPathItem):
    def __init__(self, parent):
        """The parent should be a VirtualHelixItem."""
        if self.__class__ == AbstractStrandItem:
            raise NotImplementedError("AbstractStrandItem should be subclassed.")
        super(AbstractStrandItem, self).__init__(parent)
        self._strand = None
        self._oligo = None

    ### SIGNALS ###

    ### SLOTS ###
    def oligoAppeareanceChanged(self):
        """docstring for oligoAppeareanceChanged"""
        pass

    def hasNewOligoSlot(self, oligo):
        """docstring for hasNewOligoSlot"""
        self._oligo = oligo
        # redraw

    def strandRemovedSlot(self, strand):
        """docstring for strandRemovedSlot"""
        pass

    def decoratorAddedSlot(self, decorator):
        """docstring for decoratorAddedSlot"""
        pass

    ### METHODS ###
    def connectSignals(self):
        self._oligo.appearanceChangedSignal.connect(self.oligoAppeareanceChanged)
        self._strand.strandHasNewOligoSignal.connect(self.hasNewOligoSlot)
        self._strand.destroyedSignal.connect(self.strandRemovedSlot)
        self._strand.decoratorAddedSignal.connect(self.decoratorAddedSlot)

    def disconnectSignals(self):
        self._oligo.appearanceChangedSignal.disconnect(self.oligoAppeareanceChanged)
        self._strand.strandHasNewOligoSignal.disconnect(self.hasNewOligoSlot)
        self._strand.destroyedSignal.disconnect(self.strandRemovedSlot)
        self._strand.decoratorAddedSignal.disconnect(self.decoratorAddedSlot)

    ### COMMANDS ###



"""
unittests.py

Created by Shawn Douglas on 2011-06-28.
"""

import sys, os
sys.path.insert(0, '.')

import time, code
from PyQt4.QtCore import *
from PyQt4.QtGui import *
import tests.cadnanoguitestcase
from tests.cadnanoguitestcase import CadnanoGuiTestCase
from cadnano2.model.enum import StrandType
from cadnano2.model.virtualhelix import VirtualHelix
import unittest
from rangeset import RangeSet, rangeIntersection
import random
seed = random.Random().randint(0,1<<32)
enviroseed = os.environ.get('UNITTESTS_PRNG_SEED', False)
if enviroseed != False:
    seed = int(enviroseed)
del enviroseed
print("Seeding tests.unittests; use setenv UNITTESTS_PRNG_SEED=%i to replay."%seed)


class UnitTests(CadnanoGuiTestCase):
    """
    Unit tests should test individual modules, and do not necessarily need
    to simulate user interaction.

    Create new tests by adding methods to this class that begin with "test".
    See for more detail: http://docs.python.org/library/unittest.html

    Run unit tests by calling "python -m test.unittests" from cadnano2 root
    directory.
    """
    def setUp(self):
        """
        The setUp method is called before running any test. It is used
        to set the general conditions for the tests to run correctly.
        """
        CadnanoGuiTestCase.setUp(self)
        self.prng = random.Random(seed)
        # Add extra unit-test-specific initialization here

    def tearDown(self):
        """
        The tearDown method is called at the end of running each test,
        generally used to clean up any objects created in setUp
        """
        CadnanoGuiTestCase.tearDown(self)
        # Add unit-test-specific cleanup here

    def testUnit1(self):
        """docstring for testUnit1"""
        pass

if __name__ == '__main__':
    tc = UnitTests()
    tc.setUp()
    # tc.testRangeSet_addRange_removeRange()
    tests.cadnanoguitestcase.main()



#!/usr/bin/env python
# encoding: utf-8

from heapq import heapify, heappush, heappop
from itertools import product
from collections import defaultdict
import random

from cadnano2.model.enum import StrandType
from cadnano2.model.virtualhelix import VirtualHelix
from cadnano2.model.strand import Strand
from cadnano2.model.oligo import Oligo
from cadnano2.model.strandset import StrandSet
from cadnano2.views import styles

import cadnano2.util as util

util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QUndoCommand'])


class Part(QObject):
    """
    A Part is a group of VirtualHelix items that are on the same lattice.
    Parts are the model component that most directly corresponds to a
    DNA origami design.

    Parts are always parented to the document.
    Parts know about their oligos, and the internal geometry of a part
    Copying a part recursively copies all elements in a part:
        VirtualHelices, Strands, etc

    PartInstances are parented to either the document or an assembly
    PartInstances know global position of the part
    Copying a PartInstance only creates a new PartInstance with the same
    Part(), with a mutable parent and position field.
    """

    _step = 21  # this is the period (in bases) of the part lattice
    _radius = 1.125  # nanometers
    _turnsPerStep = 2
    _helicalPitch = _step / _turnsPerStep
    _twistPerBase = 360 / _helicalPitch  # degrees

    def __init__(self, *args, **kwargs):
        """
        Sets the parent document, sets bounds for part dimensions, and sets up
        bookkeeping for partInstances, Oligos, VirtualHelix's, and helix ID
        number assignment.
        """
        if self.__class__ == Part:
            e = "This class is abstract. Perhaps you want HoneycombPart."
            raise NotImplementedError(e)
        self._document = kwargs.get('document', None)
        super(Part, self).__init__(parent=self._document)
        # Data structure
        self._insertions = defaultdict(dict)  # dict of insertions per virtualhelix
        self._oligos = set()
        self._coordToVirtualHelix = {}
        self._numberToVirtualHelix = {}
        # Dimensions
        self._maxRow = 50  # subclass overrides based on prefs
        self._maxCol = 50
        self._minBase = 0
        self._maxBase = int(2 * self._step - 1)
        # ID assignment
        self.oddRecycleBin, self.evenRecycleBin = [], []
        self.reserveBin = set()
        self._highestUsedOdd = -1  # Used in _reserveHelixIDNumber
        self._highestUsedEven = -2  # same
        self._importedVHelixOrder = None
        # Runtime state
        self._activeBaseIndex = self._step
        self._activeVirtualHelix = None
        self._activeVirtualHelixIdx = None

    # end def

    def __repr__(self):
        clsName = self.__class__.__name__
        return "<%s %s>" % (clsName, str(id(self))[-4:])

    ### SIGNALS ###
    partActiveSliceIndexSignal = pyqtSignal(QObject, int)  # self, index
    partActiveSliceResizeSignal = pyqtSignal(QObject)      # self
    partDimensionsChangedSignal = pyqtSignal(QObject)      # self
    partInstanceAddedSignal = pyqtSignal(QObject)          # self
    partParentChangedSignal = pyqtSignal(QObject)          # self
    partPreDecoratorSelectedSignal = pyqtSignal(object, int, int, int)  # row,col,idx
    partRemovedSignal = pyqtSignal(QObject)                # self
    partStrandChangedSignal = pyqtSignal(object, QObject)          # self, virtualHelix
    partVirtualHelixAddedSignal = pyqtSignal(object, QObject)      # self, virtualhelix
    partVirtualHelixRenumberedSignal = pyqtSignal(object, tuple)   # self, coord
    partVirtualHelixResizedSignal = pyqtSignal(object, tuple)      # self, coord
    partVirtualHelicesReorderedSignal = pyqtSignal(object, list)   # self, list of coords
    partHideSignal = pyqtSignal(QObject)
    partActiveVirtualHelixChangedSignal = pyqtSignal(QObject, QObject)

    ### SLOTS ###

    ### ACCESSORS ###
    def document(self):
        return self._document
    # end def

    def oligos(self):
        return self._oligos
    # end def

    def setDocument(self, document):
        self._document = document
    # end def

    def stepSize(self):
        return self._step
    # end def

    def subStepSize(self):
        """Note: _subStepSize is defined in subclasses."""
        return self._subStepSize
    # end def

    def undoStack(self):
        return self._document.undoStack()
    # end def

    ### PUBLIC METHODS FOR QUERYING THE MODEL ###
    def virtualHelix(self, vhref, returnNoneIfAbsent=True):
        # vhrefs are the shiny new way to talk to part about its constituent
        # virtualhelices. Wherever you see f(...,vhref,...) you can
        # f(...,27,...)         use the virtualhelix's id number
        # f(...,vh,...)         use an actual virtualhelix
        # f(...,(1,42),...)     use the coordinate representation of its position
        """A vhref is the number of a virtual helix, the (row, col) of a virtual helix,
        or the virtual helix itself. For conveniece, CRUD should now work with any of them."""
        vh = None
        if type(vhref) in (int, int):
            vh = self._numberToVirtualHelix.get(vhref, None)
        elif type(vhref) in (tuple, list):
            vh = self._coordToVirtualHelix.get(vhref, None)
        else:
            vh = vhref
        if not isinstance(vh, VirtualHelix):
            if returnNoneIfAbsent:
                return None
            else:
                err = "Couldn't find the virtual helix in part %s "+\
                      "referenced by index %s" % (self, vhref)
                raise IndexError(err)
        return vh

    def activeBaseIndex(self):
        return self._activeBaseIndex
    # end def

    def activeVirtualHelix(self):
        return self._activeVirtualHelix
     # end def

    def activeVirtualHelixIdx(self):
        return self._activeVirtualHelixIdx
     # end def

    def dimensions(self):
        """Returns a tuple of the max X and maxY coordinates of the lattice."""
        return self.latticeCoordToPositionXY(self._maxRow, self._maxCol)
    # end def

    def getStapleSequences(self):
        """getStapleSequences"""
        s = "Start,End,Sequence,Length,Color\n"
        for oligo in self._oligos:
            if oligo.strand5p().strandSet().isStaple():
                s = s + oligo.sequenceExport()
        return s

    def getVirtualHelices(self):
        """yield an iterator to the virtualHelix references in the part"""
        return list(self._coordToVirtualHelix.values())
    # end def

    def indexOfRightmostNonemptyBase(self):
        """
        During reduction of the number of bases in a part, the first click
        removes empty bases from the right hand side of the part (red
        left-facing arrow). This method returns the new numBases that will
        effect that reduction.
        """
        ret = self._step - 1
        for vh in self.getVirtualHelices():
            ret = max(ret, vh.indexOfRightmostNonemptyBase())
        return ret
    # end def

    def insertions(self):
        """Return dictionary of insertions."""
        return self._insertions
    # end def

    def isEvenParity(self, row, column):
        """Should be overridden when subclassing."""
        raise NotImplementedError
    # end def

    def getStapleLoopOligos(self):
        """
        Returns staple oligos with no 5'/3' ends. Used by
        actionExportStaplesSlot in documentcontroller to validate before
        exporting staple sequences.
        """
        stapLoopOlgs = []
        for o in list(self.oligos()):
            if o.isStaple() and o.isLoop():
                stapLoopOlgs.append(o)
        return stapLoopOlgs

    def hasVirtualHelixAtCoord(self, coord):
        return coord in self._coordToVirtualHelix
    # end def

    def maxBaseIdx(self):
        return self._maxBase
    # end def

    def minBaseIdx(self):
        return self._minBase
    # end def

    def numberOfVirtualHelices(self):
        return len(self._coordToVirtualHelix)
    # end def

    def radius(self):
        return self._radius
    # end def

    def helicalPitch(self):
        return self._helicalPitch
    # end def

    def twistPerBase(self):
        return self._twistPerBase
    # end def

    def virtualHelixAtCoord(self, coord):
        """
        Looks for a virtualHelix at the coordinate, coord = (row, colum)
        if it exists it is returned, else None is returned
        """
        try:
            return self._coordToVirtualHelix[coord]
        except:
            return None
    # end def

    ### PUBLIC METHODS FOR EDITING THE MODEL ###
    def autoStaple(part):
        """Autostaple does the following:
        1. Clear existing staple strands by iterating over each strand
        and calling RemoveStrandCommand on each. The next strand to remove
        is always at index 0.
        2. Create temporary strands that span regions where scaffold is present.
        3. Determine where actual strands will go based on strand overlap with
        prexovers.
        4. Delete temporary strands and create new strands.
        """
        epDict = {}  # keyed on StrandSet
        cmds = []

        # clear existing staple strands
        # part.verifyOligos()

        for o in list(part.oligos()):
            if not o.isStaple():
                continue
            c = Oligo.RemoveOligoCommand(o)
            cmds.append(c)
        # end for
        util.execCommandList(part, cmds, desc="Clear staples")
        cmds = []

        # create strands that span all bases where scaffold is present
        for vh in part.getVirtualHelices():
            segments = []
            scafSS = vh.scaffoldStrandSet()
            for strand in scafSS:
                lo, hi = strand.idxs()
                if len(segments) == 0:
                    segments.append([lo, hi])  # insert 1st strand
                elif segments[-1][1] == lo - 1:
                    segments[-1][1] = hi  # extend
                else:
                    segments.append([lo, hi])  # insert another strand
            stapSS = vh.stapleStrandSet()
            epDict[stapSS] = []
            for i in range(len(segments)):
                lo, hi = segments[i]
                epDict[stapSS].extend(segments[i])
                c = StrandSet.CreateStrandCommand(stapSS, lo, hi, i)
                cmds.append(c)
        util.execCommandList(part, cmds, desc="Add tmp strands", useUndoStack=False)
        cmds = []

        # determine where xovers should be installed
        for vh in part.getVirtualHelices():
            stapSS = vh.stapleStrandSet()
            scafSS = vh.scaffoldStrandSet()
            is5to3 = stapSS.isDrawn5to3()
            potentialXovers = part.potentialCrossoverList(vh)
            for neighborVh, idx, strandType, isLowIdx in potentialXovers:
                if strandType != StrandType.Staple:
                    continue
                if isLowIdx and is5to3:
                    strand = stapSS.getStrand(idx)
                    neighborSS = neighborVh.stapleStrandSet()
                    nStrand = neighborSS.getStrand(idx)
                    if strand == None or nStrand == None:
                        continue
                    # check for bases on both strands at [idx-1:idx+3]
                    if not (strand.lowIdx() < idx and strand.highIdx() > idx + 1):
                        continue
                    if not (nStrand.lowIdx() < idx and nStrand.highIdx() > idx + 1):
                        continue

                    # check for nearby scaffold xovers
                    scafStrandL = scafSS.getStrand(idx-4)
                    scafStrandH = scafSS.getStrand(idx+5)
                    if scafStrandL:
                        if scafStrandL.hasXoverAt(idx-4):
                            continue
                    if scafStrandH:
                        if scafStrandH.hasXoverAt(idx+5):
                            continue

                    # disable edge xovers
                    scafStrandL1 = scafSS.getStrand(idx-1)
                    scafStrandM = scafSS.getStrand(idx)
                    scafStrandH1 = scafSS.getStrand(idx+1)
                    if scafStrandL1:
                        if scafStrandL1.hasXoverAt(idx-1) and not vh.hasStrandAtIdx(idx-2):
                            continue
                        if scafStrandL1.hasXoverAt(idx-2) and not vh.hasStrandAtIdx(idx-3):
                            continue
                    if scafStrandM:
                        if scafStrandM.hasXoverAt(idx-1) and not vh.hasStrandAtIdx(idx-2):
                            continue
                        if scafStrandM.hasXoverAt(idx+1) and not vh.hasStrandAtIdx(idx+2):
                            continue
                    if scafStrandH1:
                        if scafStrandH1.hasXoverAt(idx+1) and not vh.hasStrandAtIdx(idx+2):
                            continue
                        if scafStrandH1.hasXoverAt(idx+2) and not vh.hasStrandAtIdx(idx+3):
                            continue

                    # Finally, add the xovers to install
                    epDict[stapSS].extend([idx, idx+1])
                    epDict[neighborSS].extend([idx, idx+1])

        # clear temporary staple strands
        for vh in part.getVirtualHelices():
            stapSS = vh.stapleStrandSet()
            for strand in stapSS:
                c = StrandSet.RemoveStrandCommand(stapSS, strand, 0)
                cmds.append(c)
        util.execCommandList(part, cmds, desc="Rm tmp strands", useUndoStack=False)
        cmds = []

        util.beginSuperMacro(part, desc="Auto-Staple")

        for stapSS, epList in epDict.items():
            assert (len(epList) % 2 == 0)
            epList = sorted(epList)
            ssIdx = 0
            for i in range(0, len(epList),2):
                lo, hi = epList[i:i+2]
                c = StrandSet.CreateStrandCommand(stapSS, lo, hi, ssIdx)
                cmds.append(c)
                ssIdx += 1
        util.execCommandList(part, cmds, desc="Create strands")
        cmds = []

        # create crossovers wherever possible (from strand5p only)
        for vh in part.getVirtualHelices():
            stapSS = vh.stapleStrandSet()
            is5to3 = stapSS.isDrawn5to3()
            potentialXovers = part.potentialCrossoverList(vh)
            for neighborVh, idx, strandType, isLowIdx in potentialXovers:
                if strandType != StrandType.Staple:
                    continue
                if (isLowIdx and is5to3) or (not isLowIdx and not is5to3):
                    strand = stapSS.getStrand(idx)
                    neighborSS = neighborVh.stapleStrandSet()
                    nStrand = neighborSS.getStrand(idx)
                    if strand == None or nStrand == None:
                        continue
                    if idx in strand.idxs() and idx in nStrand.idxs():
                        # only install xovers on pre-split strands
                        part.createXover(strand, idx, nStrand, idx, updateOligo=False)

        c = Part.RefreshOligosCommand(part)
        cmds.append(c)
        util.execCommandList(part, cmds, desc="Assign oligos")

        cmds = []
        util.endSuperMacro(part)

    # end def

    def verifyOligoStrandCounts(self):
        total_stap_strands = 0
        stapOligos = set()
        total_stap_oligos = 0

        for vh in self.getVirtualHelices():
            stapSS = vh.stapleStrandSet()
            total_stap_strands += len(stapSS._strandList)
            for strand in stapSS:
                stapOligos.add(strand.oligo())
        # print "# stap oligos:", len(stapOligos), "# stap strands:", total_stap_strands


    def verifyOligos(self):
        total_errors = 0
        total_passed = 0

        for o in list(self.oligos()):
            oL = o.length()
            a = 0
            gen = o.strand5p().generator3pStrand()

            for s in gen:
                a += s.totalLength()
            # end for
            if oL != a:
                total_errors += 1
                # print "wtf", total_errors, "oligoL", oL, "strandsL", a, "isStaple?", o.isStaple()
                o.applyColor('#ff0000')
            else:
                total_passed += 1
        # end for
        # print "Total Passed: ", total_passed, "/", total_passed+total_errors
    # end def

    def removeVirtualHelices(self, useUndoStack=True):
        vhs = [vh for vh in self._coordToVirtualHelix.values()]
        for vh in vhs:
            vh.remove(useUndoStack)
        # end for
    # end def

    # def remove(self, useUndoStack=True):
    #     """
    #     This method uses the slow method of removing each element one at a time
    #     while maintaining state while the command is executed
    #     """
    #     self.partHideSignal.emit(self)
    #     self._activeVirtualHelix = None
    #     if useUndoStack:
    #         self.undoStack().beginMacro("Delete Part")
    #     self.removeVirtualHelices(useUndoStack)
    #     c = Part.RemovePartCommand(self)
    #     if useUndoStack:
    #         self.undoStack().push(c)
    #         self.undoStack().endMacro()
    #     else:
    #         c.redo()
    # # end def

    def remove(self, useUndoStack=True):
        """
        This method assumes all strands are and all VirtualHelices are
        going away, so it does not maintain a valid model state while
        the command is being executed.
        Everything just gets pushed onto the undostack more or less as is.
        Except that strandSets are actually cleared then restored, but this
        is neglible performance wise.  Also, decorators/insertions are assumed
        to be parented to strands in the view so their removal Signal is
        not emitted.  This causes problems with undo and redo down the road
        but works as of now.
        """
        self.partHideSignal.emit(self)
        self._activeVirtualHelix = None
        if useUndoStack:
            self.undoStack().beginMacro("Delete Part")
        # remove strands and oligos
        self.removeAllOligos(useUndoStack)
        # remove VHs
        vhs = list(self._coordToVirtualHelix.values())
        for vh in vhs:
            d = VirtualHelix.RemoveVirtualHelixCommand(self, vh)
            if useUndoStack:
                self.undoStack().push(d)
            else:
                d.redo()
        # end for
        # remove the part
        e = Part.RemovePartCommand(self)
        if useUndoStack:
            self.undoStack().push(e)
            self.undoStack().endMacro()
        else:
            e.redo()
    # end def
    
    def removeAllOligos(self, useUndoStack=True):
        # clear existing oligos
        cmds = []
        for o in list(self.oligos()):
            cmds.append(Oligo.RemoveOligoCommand(o))
        # end for
        util.execCommandList(self, cmds, desc="Clear oligos", useUndoStack=useUndoStack)
    # end def

    def addOligo(self, oligo):
        self._oligos.add(oligo)

    # end def

    def createVirtualHelix(self, row, col, useUndoStack=True):
        c = Part.CreateVirtualHelixCommand(self, row, col)
        util.execCommandList(self, [c], desc="Add VirtualHelix", \
                                                useUndoStack=useUndoStack)
    # end def

    def createXover(self, strand5p, idx5p, strand3p, idx3p, updateOligo=True, useUndoStack=True):
        # prexoveritem needs to store left or right, and determine
        # locally whether it is from or to
        # pass that info in here in and then do the breaks
        ss5p = strand5p.strandSet()
        ss3p = strand3p.strandSet()
        if ss5p.strandType() != ss3p.strandType():
            return
        if useUndoStack:
            self.undoStack().beginMacro("Create Xover")
        if ss5p.isScaffold() and useUndoStack:  # ignore on import
            strand5p.oligo().applySequence(None)
            strand3p.oligo().applySequence(None)
        if strand5p == strand3p:
            """
            This is a complicated case basically we need a truth table.
            1 strand becomes 1, 2 or 3 strands depending on where the xover is
            to.  1 and 2 strands happen when the xover is to 1 or more existing
            endpoints.  Since SplitCommand depends on a StrandSet index, we need
            to adjust this strandset index depending which direction the crossover is
            going in.

            Below describes the 3 strand process
            1) Lookup the strands strandset index (ssIdx)
            1) Split attempted on the 3 prime strand, AKA 5prime endpoint of
            one of the new strands.  We have now created 2 strands, and the ssIdx
            is either the same as the first lookup, or one more than it depending
            on which way the the strand is drawn (isDrawn5to3).  If a split occured
            the 5prime strand is definitely part of the 3prime strand created in this step
            2) Split is attempted on the resulting 2 strands.  There is
            now 3 strands, and the final 3 prime strand may be one of the two new strands
            created in this step. Check it.
            3) Create the Xover
            """
            c = None
            # lookup the initial strandset index
            found, overlap, ssIdx3p = ss3p._findIndexOfRangeFor(strand3p)
            if strand3p.idx5Prime() == idx3p:  # yes, idx already matches
                temp5 = xoStrand3 = strand3p
            else:
                offset3p = -1 if ss3p.isDrawn5to3() else 1
                if ss3p.strandCanBeSplit(strand3p, idx3p + offset3p):
                    c = ss3p.SplitCommand(strand3p, idx3p + offset3p, ssIdx3p)
                    # cmds.append(c)
                    xoStrand3 = c._strandHigh if ss3p.isDrawn5to3() else c._strandLow
                    # adjust the target 5prime strand, always necessary if a split happens here
                    if idx5p > idx3p and ss3p.isDrawn5to3():
                        temp5 = xoStrand3
                    elif idx5p < idx3p and not ss3p.isDrawn5to3():
                        temp5 = xoStrand3
                    else:
                        temp5 = c._strandLow if ss3p.isDrawn5to3() else c._strandHigh
                    if useUndoStack:
                        self.undoStack().push(c)
                    else:
                        c.redo()
                else:
                    if useUndoStack:
                        self.undoStack().endMacro()
                        # unclear the applied sequence
                        if self.undoStack().canUndo() and ss5p.isScaffold():
                            self.undoStack().undo()
                    return
                # end if
            if xoStrand3.idx3Prime() == idx5p:
                xoStrand5 = temp5
            else:
                ssIdx5p = ssIdx3p
                # if the strand was split for the strand3p, then we need to adjust the strandset index
                if c:
                    # the insertion index into the set is increases
                    if ss3p.isDrawn5to3():
                        ssIdx5p = ssIdx3p + 1 if idx5p > idx3p else ssIdx3p
                    else:
                        ssIdx5p = ssIdx3p + 1 if idx5p > idx3p else ssIdx3p
                if ss5p.strandCanBeSplit(temp5, idx5p):
                    d = ss5p.SplitCommand(temp5, idx5p, ssIdx5p)
                    # cmds.append(d)
                    xoStrand5 = d._strandLow if ss5p.isDrawn5to3() else d._strandHigh
                    if useUndoStack:
                        self.undoStack().push(d)
                    else:
                        d.redo()
                    # adjust the target 3prime strand, IF necessary
                    if idx5p > idx3p and ss3p.isDrawn5to3():
                        xoStrand3 = xoStrand5
                    elif idx5p < idx3p and not ss3p.isDrawn5to3():
                        xoStrand3 = xoStrand5
                else:
                    if useUndoStack:
                        self.undoStack().endMacro()
                        # unclear the applied sequence
                        if self.undoStack().canUndo() and ss5p.isScaffold():
                            self.undoStack().undo()
                    return
        # end if
        else:  # Do the following if it is in fact a different strand
            # is the 5' end ready for xover installation?
            if strand3p.idx5Prime() == idx3p:  # yes, idx already matches
                xoStrand3 = strand3p
            else:  # no, let's try to split
                offset3p = -1 if ss3p.isDrawn5to3() else 1
                if ss3p.strandCanBeSplit(strand3p, idx3p + offset3p):
                    found, overlap, ssIdx = ss3p._findIndexOfRangeFor(strand3p)
                    if found:
                        c = ss3p.SplitCommand(strand3p, idx3p + offset3p, ssIdx)
                        # cmds.append(c)
                        xoStrand3 = c._strandHigh if ss3p.isDrawn5to3() else c._strandLow
                        if useUndoStack:
                            self.undoStack().push(c)
                        else:
                            c.redo()
                else:  # can't split... abort
                    if useUndoStack:
                        self.undoStack().endMacro()
                        # unclear the applied sequence
                        if self.undoStack().canUndo() and ss5p.isScaffold():
                            self.undoStack().undo()
                    return

            # is the 3' end ready for xover installation?
            if strand5p.idx3Prime() == idx5p:  # yes, idx already matches
                xoStrand5 = strand5p
            else:
                if ss5p.strandCanBeSplit(strand5p, idx5p):
                    found, overlap, ssIdx = ss5p._findIndexOfRangeFor(strand5p)
                    if found:
                        d = ss5p.SplitCommand(strand5p, idx5p, ssIdx)
                        # cmds.append(d)
                        xoStrand5 = d._strandLow if ss5p.isDrawn5to3() else d._strandHigh
                        if useUndoStack:
                            self.undoStack().push(d)
                        else:
                            d.redo()
                else:  # can't split... abort
                    if useUndoStack:
                        self.undoStack().endMacro()
                        # unclear the applied sequence
                        if self.undoStack().canUndo() and ss5p.isScaffold():
                            self.undoStack().undo()
                    return
        # end else

        e = Part.CreateXoverCommand(self, xoStrand5, idx5p, xoStrand3, idx3p, updateOligo=updateOligo)
        if useUndoStack:
            self.undoStack().push(e)
            self.undoStack().endMacro()
        else:
            e.redo()

    # end def

    def removeXover(self, strand5p, strand3p, useUndoStack=True):
        cmds = []
        if strand5p.connection3p() == strand3p:
            c = Part.RemoveXoverCommand(self, strand5p, strand3p)
            cmds.append(c)
            util.execCommandList(self, cmds, desc="Remove Xover", \
                                                    useUndoStack=useUndoStack)
    # end def

    def destroy(self):
        self.setParent(None)
        self.deleteLater()  # QObject also emits a destroyed() Signal
    # end def

    def generatorFullLattice(self):
        """
        Returns a generator that yields the row, column lattice points to draw
        relative to the part origin.
        """
        return product(list(range(self._maxRow)), list(range(self._maxCol)))
    # end def

    def generatorSpatialLattice(self, scaleFactor=1.0):
        """
        Returns a generator that yields the XY spatial lattice points to draw
        relative to the part origin.
        """
        # nested for loop in one line
        latticeCoordToPositionXY = self.latticeCoordToPositionXY
        for latticeCoord in product(list(range(self._maxRow)), list(range(self._maxCol))):
            row, col = latticeCoord
            x, y = latticeCoordToPositionXY(row, col, scaleFactor)
            yield x, y, row, col
    # end def

    def getPreXoversHigh(self, strandType, neighborType, minIdx=0, maxIdx=None):
        """
        Returns all prexover positions for neighborType that are below
        maxIdx. Used in emptyhelixitem.py.
        """
        preXO = self._scafH if strandType == StrandType.Scaffold else self._stapH
        if maxIdx == None:
            maxIdx = self._maxBase
        steps = (self._maxBase // self._step) + 1
        ret = [i * self._step + j for i in range(steps) for j in preXO[neighborType]]
        return [x for x in ret if x >= minIdx and x <= maxIdx]

    def getPreXoversLow(self, strandType, neighborType, minIdx=0, maxIdx=None):
        """
        Returns all prexover positions for neighborType that are above
        minIdx. Used in emptyhelixitem.py.
        """
        preXO = self._scafL if strandType == StrandType.Scaffold \
                                else self._stapL
        if maxIdx == None:
            maxIdx = self._maxBase
        steps = (self._maxBase // self._step) + 1
        ret = [i * self._step + j for i in range(steps) for j in preXO[neighborType]]
        return [x for x in ret if x >= minIdx and x <= maxIdx]

    def latticeCoordToPositionXY(self, row, col, scaleFactor=1.0):
        """
        Returns a tuple of the (x,y) position for a given lattice row and
        column.

        Note: The x,y position is the upperLeftCorner for the given
        coordinate, and relative to the part instance.
        """
        raise NotImplementedError  # To be implemented by Part subclass
    # end def

    def positionToCoord(self, x, y, scaleFactor=1.0):
        """
        Returns a tuple (row, column) lattice coordinate for a given
        x and y position that is within +/- 0.5 of a true valid lattice
        position.

        Note: mapping should account for int-to-float rounding errors.
        x,y is relative to the Part Instance Position.
        """
        raise NotImplementedError  # To be implemented by Part subclass
    # end def

    def newPart(self):
        return Part(self._document)
    # end def

    def removeOligo(self, oligo):
        # Not a designated method
        # (there exist methods that also directly
        # remove parts from self._oligos)
        try:
            self._oligos.remove(oligo)
        except KeyError:
            print(util.trace(5))
            # print "error removing oligo", oligo
    # end def

    def renumber(self, coordList, useUndoStack=True):
        if useUndoStack:
            self.undoStack().beginMacro("Renumber VirtualHelices")
        c = Part.RenumberVirtualHelicesCommand(self, coordList)
        if useUndoStack:
            self.undoStack().push(c)
            self.undoStack().endMacro()
        else:
            c.redo()
    # end def
    
    class RenumberVirtualHelicesCommand(QUndoCommand):
        """
        """
        def __init__(self, part, coordList):
            super(Part.RenumberVirtualHelicesCommand, self).__init__()
            self._part = part
            self._vhs = [part.virtualHelixAtCoord(coord) for coord in coordList]
            self._oldNumbers = [vh.number() for vh in self._vhs]
        # end def
            
        def redo(self):
            even = 0
            odd = 1
            for vh in self._vhs:
                if vh.isEvenParity():
                    vh.setNumber(even)
                    even += 2
                else:
                    vh.setNumber(odd)
                    odd += 2
            # end for
            part = self._part
            aVH =  part.activeVirtualHelix()
            if aVH:
                part.partStrandChangedSignal.emit(part, aVH)
            for oligo in part._oligos:
                for strand in oligo.strand5p().generator3pStrand():
                    strand.strandUpdateSignal.emit(strand)
        # end def
            
        def undo(self):
            for vh, num in zip(self._vhs, self._oldNumbers):
                vh.setNumber(num)
            # end for
            part = self._part
            aVH =  part.activeVirtualHelix()
            if aVH:
                part.partStrandChangedSignal.emit(part, aVH)
            for oligo in part._oligos:
                for strand in oligo.strand5p().generator3pStrand():
                    strand.strandUpdateSignal.emit(strand)
        # end def
    # end def

    def resizeLattice(self):
        """docstring for resizeLattice"""
        pass
    # end def

    def resizeVirtualHelices(self, minDelta, maxDelta, useUndoStack=True):
        """docstring for resizeVirtualHelices"""
        c = Part.ResizePartCommand(self, minDelta, maxDelta)
        util.execCommandList(self, [c], desc="Resize part", \
                                                    useUndoStack=useUndoStack)
    # end def

    def setActiveBaseIndex(self, idx):
        self._activeBaseIndex = idx
        self.partActiveSliceIndexSignal.emit(self, idx)
    # end def

    def setActiveVirtualHelix(self, virtualHelix, idx=None):
        self._activeVirtualHelix = virtualHelix
        self._activeVirtualHelixIdx = idx
        self.partStrandChangedSignal.emit(self, virtualHelix)
    # end def

    def selectPreDecorator(self, selectionList):
        """
        Handles view notifications that a predecorator has been selected.
        """
        if (len(selectionList) == 0):
            return
            # print "all PreDecorators were unselected"
            # partPreDecoratorUnSelectedSignal.emit()
        sel = selectionList[0]
        (row, col, baseIdx) = (sel[0], sel[1], sel[2])
        self.partPreDecoratorSelectedSignal.emit(self, row, col, baseIdx)

    def xoverSnapTo(self, strand, idx, delta):
        """
        Returns the nearest xover position to allow snap-to behavior in
        resizing strands via dragging selected xovers.
        """
        strandType = strand.strandType()
        if delta > 0:
            minIdx, maxIdx = idx - delta, idx + delta
        else:
            minIdx, maxIdx = idx + delta, idx - delta

        # determine neighbor strand and bind the appropriate prexover method
        lo, hi = strand.idxs()
        if idx == lo:
            connectedStrand = strand.connectionLow()
            preXovers = self.getPreXoversHigh
        else:
            connectedStrand = strand.connectionHigh()
            preXovers = self.getPreXoversLow
        connectedVh = connectedStrand.virtualHelix()

        # determine neighbor position, if any
        neighbors = self.getVirtualHelixNeighbors(strand.virtualHelix())
        if connectedVh in neighbors:
            neighborIdx = neighbors.index(connectedVh)
            try:
                newIdx = util.nearest(idx + delta,
                                    preXovers(strandType,
                                                neighborIdx,
                                                minIdx=minIdx,
                                                maxIdx=maxIdx)
                                    )
                return newIdx
            except ValueError:
                return None  # nearest not found in the expanded list
        else:  # no neighbor (forced xover?)... don't snap, just return
            return idx + delta

    ### PRIVATE SUPPORT METHODS ###
    def _addVirtualHelix(self, virtualHelix):
        """
        private method for adding a virtualHelix to the Parts data structure
        of virtualHelix references
        """
        self._coordToVirtualHelix[virtualHelix.coord()] = virtualHelix
    # end def

    def _removeVirtualHelix(self, virtualHelix):
        """
        private method for adding a virtualHelix to the Parts data structure
        of virtualHelix references
        """
        del self._coordToVirtualHelix[virtualHelix.coord()]
    # end def

    def _reserveHelixIDNumber(self, parityEven=True, requestedIDnum=None):
        """
        Reserves and returns a unique numerical label appropriate for a
        virtualhelix of a given parity. If a specific index is preferable
        (say, for undo/redo) it can be requested in num.
        """
        num = requestedIDnum
        if num != None:  # We are handling a request for a particular number
            assert num >= 0, int(num) == num
            # assert not num in self._numberToVirtualHelix
            if num in self.oddRecycleBin:
                self.oddRecycleBin.remove(num)
                heapify(self.oddRecycleBin)
                return num
            if num in self.evenRecycleBin:
                self.evenRecycleBin.remove(num)
                heapify(self.evenRecycleBin)
                return num
            self.reserveBin.add(num)
            return num
        # end if
        else:
            # Just find any valid index (subject to parity constraints)
            if parityEven:
                if len(self.evenRecycleBin):
                    return heappop(self.evenRecycleBin)
                else:
                    while self._highestUsedEven + 2 in self.reserveBin:
                        self._highestUsedEven += 2
                    self._highestUsedEven += 2
                    return self._highestUsedEven
            else:
                if len(self.oddRecycleBin):
                    return heappop(self.oddRecycleBin)
                else:
                    # use self._highestUsedOdd iff the recycle bin is empty
                    # and highestUsedOdd+2 is not in the reserve bin
                    while self._highestUsedOdd + 2 in self.reserveBin:
                        self._highestUsedOdd += 2
                    self._highestUsedOdd += 2
                    return self._highestUsedOdd
        # end else
    # end def

    def _recycleHelixIDNumber(self, n):
        """
        The caller's contract is to ensure that n is not used in *any* helix
        at the time of the calling of this function (or afterwards, unless
        reserveLabelForHelix returns the label again).
        """
        if n % 2 == 0:
            heappush(self.evenRecycleBin, n)
        else:
            heappush(self.oddRecycleBin, n)
    # end def

    def _splitBeforeAutoXovers(self, vh5p, vh3p, idx, useUndoStack=True):
        # prexoveritem needs to store left or right, and determine
        # locally whether it is from or to
        # pass that info in here in and then do the breaks
        ss5p = strand5p.strandSet()
        ss3p = strand3p.strandSet()
        cmds = []

        # is the 5' end ready for xover installation?
        if strand3p.idx5Prime() == idx5p:  # yes, idx already matches
            xoStrand3 = strand3p
        else:  # no, let's try to split
            offset3p = -1 if ss3p.isDrawn5to3() else 1
            if ss3p.strandCanBeSplit(strand3p, idx3p + offset3p):
                found, overlap, ssIdx = ss3p._findIndexOfRangeFor(strand3p)
                if found:
                    c = ss3p.SplitCommand(strand3p, idx3p + offset3p, ssIdx)
                    cmds.append(c)
                    xoStrand3 = c._strandHigh if ss3p.isDrawn5to3() else c._strandLow
            else:  # can't split... abort
                return

        # is the 3' end ready for xover installation?
        if strand5p.idx3Prime() == idx5p:  # yes, idx already matches
            xoStrand5 = strand5p
        else:
            if ss5p.strandCanBeSplit(strand5p, idx5p):
                found, overlap, ssIdx = ss5p._findIndexOfRangeFor(strand5p)
                if found:
                    d = ss5p.SplitCommand(strand5p, idx5p, ssIdx)
                    cmds.append(d)
                    xoStrand5 = d._strandLow if ss5p.isDrawn5to3() \
                                                else d._strandHigh
            else:  # can't split... abort
                return
        c = Part.CreateXoverCommand(self, xoStrand5, idx5p, xoStrand3, idx3p)
        cmds.append(c)
        util.execCommandList(self, cmds, desc="Create Xover", \
                                                useUndoStack=useUndoStack)
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def shallowCopy(self):
        part = self.newPart()
        part._virtualHelices = dict(self._virtualHelices)
        part._oligos = set(self._oligos)
        part._maxBase = self._maxBase
        return part
    # end def

    def deepCopy(self):
        """
        1) Create a new part
        2) copy the VirtualHelices
        3) Now you need to map the ORIGINALs Oligos onto the COPY's Oligos
        To do this you can for each Oligo in the ORIGINAL
            a) get the strand5p() of the ORIGINAL
            b) get the corresponding strand5p() in the COPY based on
                i) lookup the hash idNum of the ORIGINAL strand5p() VirtualHelix
                ii) get the StrandSet() that you created in Step 2 for the
                StrandType of the original using the hash idNum
        """
        # 1) new part
        part = self.newPart()
        for key, vhelix in self._virtualHelices:
            # 2) Copy VirtualHelix
            part._virtualHelices[key] = vhelix.deepCopy(part)
        # end for
        # 3) Copy oligos
        for oligo, val in self._oligos:
            strandGenerator = oligo.strand5p().generator3pStrand()
            strandType = oligo.strand5p().strandType()
            newOligo = oligo.deepCopy(part)
            lastStrand = None
            for strand in strandGenerator:
                idNum = strand.virtualHelix().number()
                newVHelix = part._virtualHelices[idNum]
                newStrandSet = newVHelix().getStrandSetByType(strandType)
                newStrand = strand.deepCopy(newStrandSet, newOligo)
                if lastStrand:
                    lastStrand.setConnection3p(newStrand)
                else:
                    # set the first condition
                    newOligo.setStrand5p(newStrand)
                newStrand.setConnection5p(lastStrand)
                newStrandSet.addStrand(newStrand)
                lastStrand = newStrand
            # end for
            # check loop condition
            if oligo.isLoop():
                s5p = newOligo.strand5p()
                lastStrand.set3pconnection(s5p)
                s5p.set5pconnection(lastStrand)
            # add to part
            oligo.add()
        # end for
        return part
    # end def

    def areSameOrNeighbors(self, virtualHelixA, virtualHelixB):
        """
        returns True or False
        """
        return virtualHelixB in self.getVirtualHelixNeighbors(virtualHelixA) or \
            virtualHelixA == virtualHelixB
    # end def

    def potentialCrossoverList(self, virtualHelix, idx=None):
        """
        Returns a list of tuples
            (neighborVirtualHelix, index, strandType, isLowIdx)

        where:

        neighborVirtualHelix is a virtualHelix neighbor of the arg virtualHelix
        index is the index where a potential Xover might occur
        strandType is from the enum (StrandType.Scaffold, StrandType.Staple)
        isLowIdx is whether or not it's the at the low index (left in the Path
        view) of a potential Xover site
        """
        vh = virtualHelix
        ret = []  # LUT = Look Up Table
        part = self
        # these are the list of crossover points simplified
        # they depend on whether the strandType is scaffold or staple
        # create a list of crossover points for each neighbor of the form
        # [(_scafL[i], _scafH[i], _stapL[i], _stapH[i]), ...]
        lutsNeighbor = list(
                            zip(
                                part._scafL,
                                part._scafH,
                                part._stapL,
                                part._stapH
                                )
                            )

        sTs = (StrandType.Scaffold, StrandType.Staple)
        numBases = part.maxBaseIdx()

        # create a range for the helical length dimension of the Part,
        # incrementing by the lattice step size.
        baseRange = list(range(0, numBases, part._step))

        if idx != None:
            baseRange = [x for x in baseRange if x >= idx - 3 * part._step and \
                                        x <= idx + 2 * part._step]

        if vh is None:
            return

        fromStrandSets = vh.getStrandSets()
        neighbors = self.getVirtualHelixNeighbors(vh)

        # print neighbors, lutsNeighbor
        for neighbor, lut in zip(neighbors, lutsNeighbor):
            if not neighbor:
                continue

            # now arrange again for iteration
            # (_scafL[i], _scafH[i]), (_stapL[i], _stapH[i]) )
            # so we can pair by StrandType
            lutScaf = lut[0:2]
            lutStap = lut[2:4]
            lut = (lutScaf, lutStap)

            toStrandSets = neighbor.getStrandSets()
            for fromSS, toSS, pts, st in zip(fromStrandSets, toStrandSets, lut, sTs):
                # test each period of each lattice for each StrandType
                for pt, isLowIdx in zip(pts, (True, False)):
                    for i, j in product(baseRange, pt):
                        index = i + j
                        if index < numBases:
                            if fromSS.hasNoStrandAtOrNoXover(index) and \
                                    toSS.hasNoStrandAtOrNoXover(index):
                                ret.append((neighbor, index, st, isLowIdx))
                            # end if
                        # end if
                    # end for
                # end for
            # end for
        # end for
        return ret
    # end def

    def possibleXoverAt(self, fromVirtualHelix, toVirtualHelix, strandType, idx):
        fromSS = fromVirtualHelix.getStrandSetByType(strandType)
        toSS = toVirtualHelix.getStrandSetByType(strandType)
        return fromSS.hasStrandAtAndNoXover(idx) and \
                toSS.hasStrandAtAndNoXover(idx)
    # end def

    def setImportedVHelixOrder(self, orderedCoordList):
        """Used on file import to store the order of the virtual helices."""
        self._importedVHelixOrder = orderedCoordList
        self.partVirtualHelicesReorderedSignal.emit(self, orderedCoordList)

    ### COMMANDS ###
    class CreateVirtualHelixCommand(QUndoCommand):
        def __init__(self, part, row, col):
            super(Part.CreateVirtualHelixCommand, self).__init__()
            self._part = part
            self._parityEven = part.isEvenParity(row, col)
            idNum = part._reserveHelixIDNumber(self._parityEven,
                                                requestedIDnum=None)
            self._vhelix = VirtualHelix(part, row, col, idNum)
            self._idNum = idNum
        # end def

        def redo(self):
            vh = self._vhelix
            part = self._part
            idNum = self._idNum
            vh.setPart(part)
            part._addVirtualHelix(vh)
            vh.setNumber(idNum)
            if not vh.number():
                part._reserveHelixIDNumber(self._parityEven,
                                            requestedIDnum=idNum)
            # end if
            part.partVirtualHelixAddedSignal.emit(part, vh)
            part.partActiveSliceResizeSignal.emit(part)
        # end def

        def undo(self):
            vh = self._vhelix
            part = self._part
            idNum = self._idNum
            part._removeVirtualHelix(vh)
            part._recycleHelixIDNumber(idNum)
            # clear out part references
            vh.setNumber(None)  # must come before setPart(None)
            vh.setPart(None)
            vh.virtualHelixRemovedSignal.emit(vh)
            part.partActiveSliceResizeSignal.emit(part)
        # end def
    # end class

    class CreateXoverCommand(QUndoCommand):
        """
        Creates a Xover from the 3' end of strand5p to the 5' end of strand3p
        this needs to
        1. preserve the old oligo of strand3p
        2. install the crossover
        3. apply the strand5p oligo to the strand3p
        """
        def __init__(self, part, strand5p, strand5pIdx, strand3p, strand3pIdx, updateOligo=True):
            super(Part.CreateXoverCommand, self).__init__()
            self._part = part
            self._strand5p = strand5p
            self._strand5pIdx = strand5pIdx
            self._strand3p = strand3p
            self._strand3pIdx = strand3pIdx
            self._oldOligo3p = strand3p.oligo()
            self._updateOligo = updateOligo
        # end def

        def redo(self):
            part = self._part
            strand5p = self._strand5p
            strand5pIdx = self._strand5pIdx
            strand3p = self._strand3p
            strand3pIdx = self._strand3pIdx
            olg5p = strand5p.oligo()
            oldOlg3p = self._oldOligo3p

            # 0. Deselect the involved strands
            doc = strand5p.document()
            doc.removeStrandFromSelection(strand5p)
            doc.removeStrandFromSelection(strand3p)

            if self._updateOligo:
                # Test for Loopiness
                if olg5p == strand3p.oligo():
                    olg5p.setLoop(True)
                else:
                    # 1. update preserved oligo length
                    olg5p.incrementLength(oldOlg3p.length())
                    # 2. Remove the old oligo and apply the 5' oligo to the 3' strand
                    oldOlg3p.removeFromPart()
                    for strand in strand3p.generator3pStrand():
                        # emits strandHasNewOligoSignal
                        Strand.setOligo(strand, olg5p)

            # 3. install the Xover
            strand5p.setConnection3p(strand3p)
            strand3p.setConnection5p(strand5p)

            ss5 = strand5p.strandSet()
            vh5p = ss5.virtualHelix()
            st5p = ss5.strandType()
            ss3 = strand3p.strandSet()
            vh3p = ss3.virtualHelix()
            st3p = ss3.strandType()

            part.partActiveVirtualHelixChangedSignal.emit(part, vh5p)
            # strand5p.strandXover5pChangedSignal.emit(strand5p, strand3p)
            if self._updateOligo:
                strand5p.strandUpdateSignal.emit(strand5p)
                strand3p.strandUpdateSignal.emit(strand3p)
        # end def

        def undo(self):
            part = self._part
            strand5p = self._strand5p
            strand5pIdx = self._strand5pIdx
            strand3p = self._strand3p
            strand3pIdx = self._strand3pIdx
            oldOlg3p = self._oldOligo3p
            olg5p = strand5p.oligo()

            # 0. Deselect the involved strands
            doc = strand5p.document()
            doc.removeStrandFromSelection(strand5p)
            doc.removeStrandFromSelection(strand3p)

            # 1. uninstall the Xover
            strand5p.setConnection3p(None)
            strand3p.setConnection5p(None)

            if self._updateOligo:
                # Test Loopiness
                if oldOlg3p.isLoop():
                    oldOlg3p.setLoop(False)
                else:
                    # 2. restore the modified oligo length
                    olg5p.decrementLength(oldOlg3p.length())
                    # 3. apply the old oligo to strand3p
                    oldOlg3p.addToPart(part)
                    for strand in strand3p.generator3pStrand():
                        # emits strandHasNewOligoSignal
                        Strand.setOligo(strand, oldOlg3p)

            ss5 = strand5p.strandSet()
            vh5p = ss5.virtualHelix()
            st5p = ss5.strandType()
            ss3 = strand3p.strandSet()
            vh3p = ss3.virtualHelix()
            st3p = ss3.strandType()

            part.partActiveVirtualHelixChangedSignal.emit(part, vh5p)
            # strand5p.strandXover5pChangedSignal.emit(strand5p, strand3p)
            if self._updateOligo:
                strand5p.strandUpdateSignal.emit(strand5p)
                strand3p.strandUpdateSignal.emit(strand3p)
        # end def
    # end class

    class RefreshOligosCommand(QUndoCommand):
        """
        RefreshOligosCommand is a post-processing step for AutoStaple.

        Normally when an xover is created, all strands in the 3' direction are
        assigned the oligo of the 5' strand. This becomes very expensive
        during autoStaple, because the Nth xover requires updating up to N-1
        strands.

        Hence, we disable oligo assignment during the xover creation step,
        and then do it all in one pass at the end with this command.
        """
        def __init__(self, part):
            super(Part.RefreshOligosCommand, self).__init__()
            self._part = part
        # end def

        def redo(self):
            visited = {}
            for vh in self._part.getVirtualHelices():
                stapSS = vh.stapleStrandSet()
                for strand in stapSS:
                    visited[strand] = False

            for strand in list(visited.keys()):
                if visited[strand]:
                    continue
                visited[strand] = True
                startOligo = strand.oligo()
                strand5gen = strand.generator5pStrand()
                # this gets the oligo and burns a strand in the generator
                strand5 = next(strand5gen)
                for strand5 in strand5gen:
                    oligo5 = strand5.oligo()
                    if oligo5 != startOligo:
                        oligo5.removeFromPart()
                        Strand.setOligo(strand5, startOligo)  # emits strandHasNewOligoSignal
                    visited[strand5] = True
                # end for
                startOligo.setStrand5p(strand5)
                # is it a loop?
                if strand.connection3p() == strand5:
                    startOligo.setLoop(True)
                else:
                    strand3gen = strand.generator3pStrand()
                    strand3 = next(strand3gen)   # burn one
                    for strand3 in strand3gen:
                        oligo3 = strand3.oligo()
                        if oligo3 != startOligo:
                            oligo3.removeFromPart()
                            Strand.setOligo(strand3, startOligo)  # emits strandHasNewOligoSignal
                        visited[strand3] = True
                    # end for
                startOligo.refreshLength()
            # end for

            oligoSet = set()
            for strand in list(visited.keys()):
                oligoSet.add(strand.oligo())
                strand.strandUpdateSignal.emit(strand)
        # end def

        def undo(self):
            """Doesn't reassign """
            pass
        # end def
    # end class

    class RemoveXoverCommand(QUndoCommand):
        """
        Removes a Xover from the 3' end of strand5p to the 5' end of strand3p
        this needs to
        1. preserve the old oligo of strand3p
        2. install the crossover
        3. update the oligo length
        4. apply the new strand3p oligo to the strand3p
        """
        def __init__(self, part, strand5p, strand3p):
            super(Part.RemoveXoverCommand, self).__init__()
            self._part = part
            self._strand5p = strand5p
            self._strand5pIdx = strand5p.idx3Prime()
            self._strand3p = strand3p
            self._strand3pIdx = strand3p.idx5Prime()
            nO3p = self._newOligo3p = strand3p.oligo().shallowCopy()
            colorList = styles.stapColors if strand5p.strandSet().isStaple() \
                                            else styles.scafColors
            nO3p.setColor(random.choice(colorList).name())
            nO3p.setLength(0)
            for strand in strand3p.generator3pStrand():
                nO3p.incrementLength(strand.totalLength())
            # end def
            nO3p.setStrand5p(strand3p)
            
            self._isLoop = strand3p.oligo().isLoop()
        # end def

        def redo(self):
            part = self._part
            strand5p = self._strand5p
            strand5pIdx = self._strand5pIdx
            strand3p = self._strand3p
            strand3pIdx = self._strand3pIdx
            newOlg3p = self._newOligo3p
            olg5p = self._strand5p.oligo()

            # 0. Deselect the involved strands
            doc = strand5p.document()
            doc.removeStrandFromSelection(strand5p)
            doc.removeStrandFromSelection(strand3p)

            # 1. uninstall the Xover
            strand5p.setConnection3p(None)
            strand3p.setConnection5p(None)

            if self._isLoop:
                olg5p.setLoop(False)
                olg5p.setStrand5p(strand3p)
            else:
                # 2. restore the modified oligo length
                olg5p.decrementLength(newOlg3p.length())
                # 3. apply the old oligo to strand3p
                newOlg3p.addToPart(part)
                for strand in strand3p.generator3pStrand():
                    # emits strandHasNewOligoSignal
                    Strand.setOligo(strand, newOlg3p)

            ss5 = strand5p.strandSet()
            vh5p = ss5.virtualHelix()
            st5p = ss5.strandType()
            ss3 = strand3p.strandSet()
            vh3p = ss3.virtualHelix()
            st3p = ss3.strandType()

            part.partActiveVirtualHelixChangedSignal.emit(part, vh5p)
            # strand5p.strandXover5pChangedSignal.emit(strand5p, strand3p)
            strand5p.strandUpdateSignal.emit(strand5p)
            strand3p.strandUpdateSignal.emit(strand3p)
        # end def

        def undo(self):
            part = self._part
            strand5p = self._strand5p
            strand5pIdx = self._strand5pIdx
            strand3p = self._strand3p
            strand3pIdx = self._strand3pIdx
            olg5p = strand5p.oligo()
            newOlg3p = self._newOligo3p

            # 0. Deselect the involved strands
            doc = strand5p.document()
            doc.removeStrandFromSelection(strand5p)
            doc.removeStrandFromSelection(strand3p)

            if self._isLoop:
                olg5p.setLoop(True)
                # No need to restore whatever the old Oligo._strand5p was
            else:
                # 1. update preserved oligo length
                olg5p.incrementLength(newOlg3p.length())
                # 2. Remove the old oligo and apply the 5' oligo to the 3' strand
                newOlg3p.removeFromPart()
                for strand in strand3p.generator3pStrand():
                    # emits strandHasNewOligoSignal
                    Strand.setOligo(strand, olg5p)
            # end else

            # 3. install the Xover
            strand5p.setConnection3p(strand3p)
            strand3p.setConnection5p(strand5p)

            ss5 = strand5p.strandSet()
            vh5p = ss5.virtualHelix()
            st5p = ss5.strandType()
            ss3 = strand3p.strandSet()
            vh3p = ss3.virtualHelix()
            st3p = ss3.strandType()

            part.partActiveVirtualHelixChangedSignal.emit(part, vh5p)
            # strand5p.strandXover5pChangedSignal.emit(strand5p, strand3p)
            strand5p.strandUpdateSignal.emit(strand5p)
            strand3p.strandUpdateSignal.emit(strand3p)
        # end def
    # end class

    class RemovePartCommand(QUndoCommand):
        """
        RemovePartCommand deletes a part. Emits partRemovedSignal.
        """
        def __init__(self, part):
            super(Part.RemovePartCommand, self).__init__()
            self._part = part
            self._doc = part.document()
        # end def

        def redo(self):
            # Remove the strand
            part = self._part
            doc = self._doc
            doc.removePart(part)
            part.setDocument(None)
            part.partRemovedSignal.emit(part)
        # end def

        def undo(self):
            part = self._part
            doc = self._doc
            doc._addPart(part)
            part.setDocument(doc)
            doc.documentPartAddedSignal.emit(doc, part)
        # end def
    # end class

    class RemoveAllStrandsCommand(QUndoCommand):
        """
        1. Remove all strands. Emits strandRemovedSignal for each.
        2. Remove all oligos. 
        """
        def __init__(self, part):
            super(Part.RemoveAllStrandsCommand, self).__init__()
            self._part = part
            self._vhs = vhs = part.getVirtualHelices()
            self._strandSets = []
            for vh in self._vhs:
                x = vh.getStrandSets()
                self._strandSets.append(x[0])
                self._strandSets.append(x[1])
            self._strandSetListCopies = \
                        [[y for y in x._strandList] for x in self._strandSets]
            self._oligos = set(part.oligos())
        # end def

        def redo(self):
            part = self._part
            # Remove the strand
            for sSet in self._strandSets:
                sList = sSet._strandList
                for strand in sList:
                    sSet.removeStrand(strand)
                # end for
                sSet._strandList = []
            #end for
            for vh in self._vhs:
                # for updating the Slice View displayed helices
                part.partStrandChangedSignal.emit(part, vh)
            # end for
            self._oligos.clear()
        # end def

        def undo(self):
            part = self._part
            # Remove the strand
            sListCopyIterator = iter(self._strandSetListCopies)
            for sSet in self._strandSets:
                sList = next(sListCopyIterator)
                for strand in sList:
                    sSet.strandsetStrandAddedSignal.emit(sSet, strand)
                # end for
                sSet._strandList = sList
            #end for
            for vh in self._vhs:
                # for updating the Slice View displayed helices
                part.partStrandChangedSignal.emit(part, vh)
            # end for
            for olg in self._oligos:
                part.addOligo(olg)
        # end def
    # end class

    class ResizePartCommand(QUndoCommand):
        """
        set the maximum and mininum base index in the helical direction

        need to adjust all subelements in the event of a change in the
        minimum index
        """
        def __init__(self, part, minHelixDelta, maxHelixDelta):
            super(Part.ResizePartCommand, self).__init__()
            self._part = part
            self._minDelta = minHelixDelta
            self._maxDelta = maxHelixDelta
            self._oldActiveIdx = part.activeBaseIndex()
        # end def

        def redo(self):
            part = self._part
            part._minBase += self._minDelta
            part._maxBase += self._maxDelta
            if self._minDelta != 0:
                self.deltaMinDimension(part, self._minDelta)
            for vh in part._coordToVirtualHelix.values():
                part.partVirtualHelixResizedSignal.emit(part, vh.coord())
            if self._oldActiveIdx > part._maxBase:
                part.setActiveBaseIndex(part._maxBase)
            part.partDimensionsChangedSignal.emit(part)
        # end def

        def undo(self):
            part = self._part
            part._minBase -= self._minDelta
            part._maxBase -= self._maxDelta
            if self._minDelta != 0:
                self.deltaMinDimension(part, self._minDelta)
            for vh in part._coordToVirtualHelix.values():
                part.partVirtualHelixResizedSignal.emit(part, vh.coord())
            if self._oldActiveIdx != part.activeBaseIndex():
                part.setActiveBaseIndex(self._oldActiveIdx)
            part.partDimensionsChangedSignal.emit(part)
        # end def

        def deltaMinDimension(self, part, minDimensionDelta):
            """
            Need to update:
            strands
            insertions
            """
            for vhDict in part._insertions.values():
                for insertion in vhDict:
                    insertion.updateIdx(minDimensionDelta)
                # end for
            # end for
            for vh in part._coordToVirtualHelix.values():
                for strand in vh.scaffoldStrand().generatorStrand():
                    strand.updateIdxs(minDimensionDelta)
                for strand in vh.stapleStrand().generatorStrand():
                    strand.updateIdxs(minDimensionDelta)
            # end for
        # end def
    # end class
# end class



#!/usr/bin/env python
# encoding: utf-8

from .strandset import StrandSet
import cadnano2.util as util
from .enum import StrandType

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QUndoStack', 'QUndoCommand'])


class VirtualHelix(QObject):
    """
    VirtualHelix is a container class for two StrandSet objects (one scaffold
    and one staple). The Strands all share the same helix axis. It is called
    "virtual" because many different Strands (i.e. sub-oligos) combine to
    form the "helix", just as many fibers may be braided together to 
    form a length of rope.
    """

    def __init__(self, part, row, col, idnum=0):
        super(VirtualHelix, self).__init__(part)
        self._coord = (row, col) # col, row
        self._part = part
        self._doc = part.document()
        self._scafStrandSet = StrandSet(StrandType.Scaffold, self)
        self._stapStrandSet = StrandSet(StrandType.Staple, self)
        # If self._part exists, it owns self._number
        # in that only it may modify it through the
        # private interface. The public interface for
        # setNumber just routes the call to the parent
        # dnapart if one is present. If self._part == None
        # the virtualhelix owns self._number and may modify it.
        self._number = None
        self.setNumber(idnum)
    # end def

    def __repr__(self):
        return "<%s(%d)>" % (self.__class__.__name__, self._number)

    ### SIGNALS ###
    virtualHelixRemovedSignal = pyqtSignal(QObject)  # self
    virtualHelixNumberChangedSignal = pyqtSignal(QObject, int)  # self, num

    ### SLOTS ###

    ### ACCESSORS ###
    def scaf(self, idx):
        """ Returns the strand at idx in self's scaffold, if any """
        return self._scafStrandSet.getStrand(idx)

    def stap(self, idx):
        """ Returns the strand at idx in self's scaffold, if any """
        return self._stapStrandSet.getStrand(idx)

    def coord(self):
        return self._coord
    # end def

    def number(self):
        return self._number
    # end def

    def part(self):
        return self._part
    # end def
    
    def document(self):
        return self._doc
    # end def
    
    def setNumber(self, number):
        if self._number != number:
            numToVhDict = self._part._numberToVirtualHelix
            numToVhDict[self._number] = None
            self._number = number
            self.virtualHelixNumberChangedSignal.emit(self, number)
            numToVhDict[number] = self
    # end def

    def setPart(self, newPart):
        self._part = newPart
        self.setParent(newPart)
    # end def

    def scaffoldStrandSet(self):
        return self._scafStrandSet
    # end def

    def stapleStrandSet(self):
        return self._stapStrandSet
    # end def

    def undoStack(self):
        return self._part.undoStack()
    # end def

    ### METHODS FOR QUERYING THE MODEL ###
    def scaffoldIsOnTop(self):
        return self.isEvenParity()

    def getStrandSetByIdx(self, idx):
        """
        This is a path-view-specific accessor
        idx == 0 means top strand
        idx == 1 means bottom strand
        """
        if idx == 0:
            if self.isEvenParity():
                return self._scafStrandSet
            else:
                return self._stapStrandSet
        else:
            if self.isEvenParity():
                return self._stapStrandSet
            else:
                return self._scafStrandSet
    # end def

    def getStrandSetByType(self, strandType):
        if strandType == StrandType.Scaffold:
            return self._scafStrandSet
        else:
            return self._stapStrandSet
    # end def

    def getStrandSets(self):
        """Return a tuple of the scaffold and staple StrandSets."""
        return self._scafStrandSet, self._stapStrandSet
    # end def

    def hasStrandAtIdx(self, idx):
        return self._scafStrandSet.hasStrandAt(idx, idx)
    # end def

    def indexOfRightmostNonemptyBase(self):
        """Returns the rightmost nonempty base in either scaf of stap."""
        return max(self._scafStrandSet.indexOfRightmostNonemptyBase(),\
                   self._stapStrandSet.indexOfRightmostNonemptyBase())
    # end def

    def isDrawn5to3(self, strandSet):
        isScaf = strandSet == self._scafStrandSet
        isEven = self.isEvenParity()
        return isEven == isScaf
    # end def

    def isEvenParity(self):
        return self._part.isEvenParity(*self._coord)
    # end def

    def strandSetBounds(self, indexHelix, indexType):
        """
        forwards the query to the strandSet
        """
        return self.strandSet(indexHelix, indexType).bounds()
    # end def

    ### METHODS FOR EDITING THE MODEL ###
    def destroy(self):
        # QObject also emits a destroyed() Signal
        self.setParent(None)
        self.deleteLater()
    # end def
    
    def remove(self, useUndoStack=True):
        """
        Removes a VirtualHelix from the model. Accepts a reference to the 
        VirtualHelix, or a (row,col) lattice coordinate to perform a lookup.
        """
        if useUndoStack:
            self.undoStack().beginMacro("Delete VirtualHelix")
        self._scafStrandSet.remove(useUndoStack)
        self._stapStrandSet.remove(useUndoStack)
        c = VirtualHelix.RemoveVirtualHelixCommand(self.part(), self)
        if useUndoStack:
            self.undoStack().push(c)
            self.undoStack().endMacro()
        else:
            c.redo()
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def deepCopy(self, part):
        """
        This only copies as deep as the VirtualHelix
        strands get copied at the oligo and added to the Virtual Helix
        """
        vh = VirtualHelix(part, self._number)
        vh._coords = (self._coord[0], self._coord[1])
        # If self._part exists, it owns self._number
        # in that only it may modify it through the
        # private interface. The public interface for
        # setNumber just routes the call to the parent
        # dnapart if one is present. If self._part == None
        # the virtualhelix owns self._number and may modify it.
        self._number = idnum
    # end def

    def getLegacyStrandSetArray(self, strandType):
        """Called by legacyencoder."""
        if strandType == StrandType.Scaffold:
            return self._scafStrandSet.getLegacyArray()
        else:
            return self._stapStrandSet.getLegacyArray()

    def shallowCopy(self):
        pass
    # end def

    # def translateCoords(self, deltaCoords):
    #     """
    #     for expanding a helix
    #     """
    #     deltaRow, deltaCol = deltaCoords
    #     row, col = self._coord
    #     self._coord = row + deltaRow, col + deltaCol
    # # end def

    class RemoveVirtualHelixCommand(QUndoCommand):
        """Inserts strandToAdd into strandList at index idx."""
        def __init__(self, part, virtualHelix):
            super(VirtualHelix.RemoveVirtualHelixCommand, self).__init__()
            self._part = part
            self._vhelix = virtualHelix
            self._idNum = virtualHelix.number()
            # is the number even or odd?  Assumes a valid idNum, row,col combo
            self._parityEven = (self._idNum % 2) == 0
            
        # end def

        def redo(self):
            vh = self._vhelix
            part = self._part
            idNum = self._idNum
            
            part._removeVirtualHelix(vh)
            part._recycleHelixIDNumber(idNum)
            # clear out part references
            vh.virtualHelixRemovedSignal.emit(vh)
            part.partActiveSliceResizeSignal.emit(part)
            # vh.setPart(None)
            # vh.setNumber(None)
        # end def

        def undo(self):
            vh = self._vhelix
            part = self._part
            idNum = self._idNum
            
            vh.setPart(part)
            part._addVirtualHelix(vh)
            # vh.setNumber(idNum)
            if not vh.number():
                part._reserveHelixIDNumber(self._parityEven, requestedIDnum=idNum)
            part.partVirtualHelixAddedSignal.emit(part, vh)
            part.partActiveSliceResizeSignal.emit(part)
        # end def
    # end class


from .abstractpathtool import AbstractPathTool
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), [])
util.qtWrapImport('QtGui', globals(), [])


class SkipTool(AbstractPathTool):
    """
    docstring for SkipTool
    """
    def __init__(self, controller):
        super(SkipTool, self).__init__(controller)

    def __repr__(self):
        return "skipTool"  # first letter should be lowercase


# from PyQt4.QtCore import *
# from PyQt4.QtGui import *
import cadnano
import tests.guitestcase

main = tests.guitestcase.main


class CadnanoGuiTestCase(tests.guitestcase.GUITestCase):
    """
    SEE: http://docs.python.org/library/unittest.html
    """
    def setUp(self):
        """
        The setUp method is called before running any test. It is used
        to set the general conditions for the tests to run correctly.
        For GUI Tests, you always have to call setWidget to tell the
        framework what you will be testing.
        """
        import sys

        self.app = cadnano.initAppWithGui() # kick off a Gui style app
        self.documentController = list(self.app.documentControllers)[0]
        self.mainWindow = self.documentController.win

        # Include this or the automatic build will hang
        self.app.dontAskAndJustDiscardUnsavedChanges = True

        # By setting the widget to the main window we can traverse and
        # interact with any part of it. Also, tearDown will close
        # the application so we don't need to worry about that.
        self.setWidget(self.mainWindow, False, None)

    def tearDown(self):
        """
        The tearDown method is called at the end of running each test,
        generally used to clean up any objects created in setUp
        """
        tests.guitestcase.GUITestCase.tearDown(self)

if __name__ == '__main__':
    tests.guitestcase.main()



"""
This file houses the code that allows the opening of cadnano files by dragging
to the icon in the dock or double clicking on the icon.
"""

import objc, os
from Foundation import *
from AppKit import *
from controllers.documentcontroller import DocumentController
from cadnano2.model.io.decoder import decode
from cadnano2.cadnano import app as sharedCadnanoObj


class CNApplicationDelegate(NSObject):
    def application_openFile_(self, app, f):
        if f == "main.py":  # ignore
            return
        extension = os.path.splitext(f)[1].lower()
        if extension not in ('.nno', '.json', '.cadnano'):
            print("Could not open file %s (bad extension %s)"%(f, extension))
            return
        dc = list(sharedCadnanoObj().documentControllers)[0]
        decode(dc.document(), file(str(f)).read())
        return None

    def application_openFiles_(self, app, fs):
        if fs.isKindOfClass_(NSCFString):
            self.application_openFile_(app, fs)
            return
        for f in fs:
            self.application_openFiles_(app, f)

    def applicationShouldTerminate_(self, app):
        for dc in sharedCadnanoObj().documentControllers:
            if not dc.maybeSave():
                return NSTerminateCancel
        return NSTerminateNow

sharedDelegate = CNApplicationDelegate.alloc().init()
NSApplication.sharedApplication().setDelegate_(sharedDelegate)



#!/usr/bin/env python
# encoding: utf-8

"""
CustomQGraphicsView.py
.. module:: CustomQGraphicsView
   :platform: Unix, Windows, Mac OS X
   :synopsis: A Custom QGraphicsView module to allow focus input events
   like mouse clicks and panning and zooming

.. moduleauthor::  Nick Conway on 2011-01-17.
Copyright (c) 2010 . All rights reserved.

"""
from cadnano2.cadnano import app
from cadnano2.views import styles

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['Qt', 'QTimer', 'pyqtSignal', 'QTimeLine'])
util.qtWrapImport('QtGui', globals(),  ['QGuiApplication', 'QPaintEngine'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsView',
                                           'QGraphicsScene'])

# for OpenGL mode
try:
    # from OpenGL import GL
    from PyQt6.QtWidgets import QOpenGLWidget
except ImportError:
    GL = False

GL = False


class CustomQGraphicsView(QGraphicsView):
    """
    Base class for QGraphicsViews with Mouse Zoom and Pan support via the
    Control/Command shortcut key.

    A QGraphics View stores info on the view and handles mouse events for
    zooming and panning

    Ctrl-MidMouseButton = Pan
    Ctrl-RightMouseButton = Dolly Zoom
    MouseWheel = Zoom

    Parameters
    ----------
    parent: type of QWidget, such as QWidget.splitter() for the type of
    View its has

    See Also
    --------

    Examples
    --------

    For details on these and other miscellaneous methods, see below.
    """
    def __init__(self, parent=None):
        """
        On initialization, we need to bind the Ctrl/command key to
        enable manipulation of the view.
        """
        QGraphicsView.__init__(self, parent)
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.setRubberBandSelectionMode(Qt.ItemSelectionMode.IntersectsItemShape)
        self.setStyleSheet("QGraphicsView { background-color: rgb(96.5%, 96.5%, 96.5%); }")
        self._noDrag = QGraphicsView.DragMode.RubberBandDrag
        self._yesDrag = QGraphicsView.DragMode.ScrollHandDrag

        # reset things that are state dependent
        self.clearGraphicsView()
        self.setAcceptDrops(True)

        self._x0 = 0
        self._y0 = 0
        self._scale_size = 1.0
        self._scale_limit_max = 3.0
        self._scale_limit_min = 0.41
        self._scaleUpRate = 0.01
        self._scaleDownRate = 0.01
        self._scaleFitFactor = 1  # sets initial zoom level
        self._showDetails = True
        self._last_scale_factor = 0.0
        self.sceneRootItem = None  # the item to transform
        # Keyboard panning
        self._key_pan_delta_x = styles.PATH_BASE_WIDTH * 21
        self._key_pan_delta_y = styles.PATH_HELIX_HEIGHT + styles.PATH_HELIX_PADDING/2
        # Modifier keys and buttons
        self._key_mod = Qt.Key.Key_Control
        self._button_pan = Qt.MouseButton.LeftButton
        self._button_pan_alt = Qt.MouseButton.MiddleButton
        self._button_zoom = Qt.MouseButton.RightButton

        self.toolbar = None  # custom hack for the paint tool palette
        self._name = None

        if GL:
            self.setViewport(QGLWidget(QGLFormat(QGL.SampleBuffers)))
            self.setViewportUpdateMode(QGraphicsView.FullViewportUpdate)
        else:
            self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.MinimalViewportUpdate)
            # self.setViewportUpdateMode(QGraphicsView.SmartViewportUpdate)
    # end def

    levelOfDetailChangedSignal = pyqtSignal(bool)

    def __repr__(self):
        clsName = self.__class__.__name__
        objId = self._name if self._name else str(id(self))[-4:]
        return "<%s %s>" % (clsName, objId)

    def setName(self, name):
        self._name = name
    # end def

    def setViewportUpdateOn(self, isEnabled):
        if isEnabled:
            self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.MinimalViewportUpdate)
        else:
            self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.NoViewportUpdate)
    # end def

    def activateSelection(self, isActive):
        if self._selectionLock:
            self._selectionLock.clearSelection(False)
        self.clearSelectionLockAndCallbacks()
        if isActive:
            self._noDrag = QGraphicsView.DragMode.RubberBandDrag
        else:
            self._noDrag = QGraphicsView.DragMode.NoDrag
        if self.dragMode() != self._yesDrag:
            self.setDragMode(self._noDrag)
    # end def

    def clearGraphicsView(self):
        # Event handling
        self._hasFocus = False
        # Misc
        self.clearSelectionLockAndCallbacks()
        # Pan and dolly defaults
        self._transformEnable = False
        self._dollyZoomEnable = False
        self.setDragMode(self._noDrag)
    # end def

    def clearSelectionLockAndCallbacks(self):
        self._selectionLock = None # a selection group to limit types of items selected
        self._pressList = [] # bookkeeping to handle passing mouseReleaseEvents to QGraphicsItems that don't get them
    # end def

    def setGLView(self, boolval):
        scene = self.scene()
        if boolval and self.isGL == False:
            self.isGL = True
            # scene.drawBackground = self.drawBackgroundGL
            # self.setViewport(QGLWidget(QGLFormat(QGL.SampleBuffers)))
            # self.setViewportUpdateMode(QGraphicsView.FullViewportUpdate)
        elif not boolval and self.isGL == True:
            self.isGL = False
            # scene.drawBackground = self.drawBackgroundNonGL
            # self.setViewport(QWidget())
            # self.setViewportUpdateMode(QGraphicsView.MinimalViewportUpdate)
    # end def

    def setupGL(self):
        scene = self.scene()
        win = self.sceneRootItem.window()
        self.isGL = True
        self.isGLSwitchAllowed = True
        self.qTimer = QTimer()
        # self.drawBackgroundNonGL = scene.drawBackground
        # scene.drawBackground = self.drawBackgroundGL
        # format = QGLFormat(QGL.SampleBuffers)
        # format.setSamples(16)
        # print "# of samples", format.samples(), format.sampleBuffers()
        # self.setViewport(QGLWidget(format))
        # self.setViewportUpdateMode(QGraphicsView.FullViewportUpdate)
    # end def

    def resetGL(self):
        scaleFactor = self.transform().m11()
        # print "scaleFactor", scaleFactor
        self.sceneRootItem.window().statusBar().showMessage("%f" % scaleFactor)

        if scaleFactor < .15:# and self.isGLSwitchAllowed:
            # self.isGLSwitchAllowed = False
            self.setGLView(True)
            self._showDetails = False
            self.levelOfDetailChangedSignal.emit(False) # zoomed out
            self.qTimer.singleShot(500, self.allowGLSwitch)
        elif scaleFactor > .2:# and self.isGLSwitchAllowed:
            # self.isGLSwitchAllowed = False
            self.setGLView(False)
            self._showDetails = True
            self.levelOfDetailChangedSignal.emit(True) # zoomed in
            self.qTimer.singleShot(500, self.allowGLSwitch)
    # end def

    def shouldShowDetails(self):
        return self._showDetails
    # end def

    def allowGLSwitch(self):
        self.isGLSwitchAllowed = True
    # end def

    def drawBackgroundGL(self, painter, rect):
        """
        This method is for overloading the QGraphicsScene.
        """
        if painter.paintEngine().type() != QPaintEngine.OpenGL and \
            painter.paintEngine().type() != QPaintEngine.OpenGL2:

            qWarning("OpenGLScene: drawBackground needs a QGLWidget to be set as viewport on the graphics view");
            return
        # end if
        painter.beginNativePainting()
        GL.glDisable(GL.GL_DEPTH_TEST) # disable for 2D drawing
        GL.glClearColor(1.0, 1.0, 1.0, 1.0)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        painter.endNativePainting()
    # end def

    def focusInEvent(self, event):
        self._hasFocus = True

    def focusOutEvent(self, event):
        self._transformEnable = False
        self._dollyZoomEnable = False
        self._hasFocus = False
        self._transformEnable = False
    # end def

    def setSelectionLock(self, selectionLock):
        self._selectionLock = selectionLock
    # end def

    def selectionLock(self):
        return self._selectionLock
    # end def

    def setScaleFitFactor(self, value):
        """docstring for setScaleFitFactor"""
        self._scaleFitFactor = value
    # end def

    def setKeyPan(self, button):
        """Set the class pan button remotely"""
        self._button_pan = button
    # end def

    def addToPressList(self, item):
        """docstring for addToPressList"""
        # self._pressList[self._pressListIdx].append(item)
        self._pressList.append(item)
    # end def

    def keyPanDeltaX(self):
        """Returns the distance in scene space to move the sceneRootItem when
        panning left or right."""
        # PyQt isn't aware that QGraphicsObject isa QGraphicsItem and so
        # it returns a separate python object if, say, childItems() returns
        # a QGraphicsObject casted to a QGraphicsItem. If this is the case,
        # we can still find the QGraphicsObject thusly:
        candidateDxDeciders = list(self.sceneRootItem.childItems())
        candidateDxDeciders = candidateDxDeciders +\
                           [cd.toGraphicsObject() for cd in candidateDxDeciders]
        for cd in candidateDxDeciders:
            if cd == None:
                continue
            keyPanDXMethod = getattr(cd, 'keyPanDeltaX', None)
            if keyPanDXMethod != None:
                return keyPanDXMethod()
        return 100

    def keyPanDeltaY(self):
        """Returns the distance in scene space to move the sceneRootItem when
        panning left or right."""
        candidateDyDeciders = list(self.sceneRootItem.childItems())
        candidateDyDeciders = candidateDyDeciders +\
                           [cd.toGraphicsObject() for cd in candidateDyDeciders]
        for cd in candidateDyDeciders:
            if cd == None:
                continue
            keyPanDYMethod = getattr(cd, 'keyPanDeltaY', None)
            if keyPanDYMethod != None:
                return keyPanDYMethod()
        return 100

    def keyPressEvent(self, event):
        """
        Handle key presses for mouse-drag transforms and arrow-key panning.
        """
        if not self._hasFocus:  # we don't have focus -> ignore keypress
            return
        if event.key() == self._key_mod:
            self._transformEnable = True
            QGraphicsView.keyPressEvent(self, event)
        elif event.key() == Qt.Key.Key_Left:
            transform = self.sceneRootItem.transform()
            transform.translate(self.keyPanDeltaX(), 0)
            self.sceneRootItem.setTransform(transform)
        elif event.key() == Qt.Key.Key_Up:
            transform = self.sceneRootItem.transform()
            transform.translate(0, self.keyPanDeltaY())
            self.sceneRootItem.setTransform(transform)
        elif event.key() == Qt.Key.Key_Right:
            transform = self.sceneRootItem.transform()
            transform.translate(-self.keyPanDeltaX(), 0)
            self.sceneRootItem.setTransform(transform)
        elif event.key() == Qt.Key.Key_Down:
            transform = self.sceneRootItem.transform()
            transform.translate(0, -self.keyPanDeltaY())
            self.sceneRootItem.setTransform(transform)
        elif event.key() == Qt.Key.Key_Plus:
            self.zoomIn(0.3)
        elif event.key() == Qt.Key.Key_Minus:
            self.zoomIn(0.03)
        else:
            return QGraphicsView.keyPressEvent(self, event)
        # end else
    # end def

    def keyReleaseEvent(self, event):
        """docstring for keyReleaseEvent"""
        if event.key() == self._key_mod:
            self._transformEnable = False
            self._dollyZoomEnable = False
            self._panDisable()
        # end if
        else:
            QGraphicsView.keyReleaseEvent(self, event)
        # end else
    # end def

    def enterEvent(self, event):
        self.setFocus()
        self.setDragMode(self._noDrag)
        QGraphicsView.enterEvent(self, event)

    def leaveEvent(self, event):
        self.clearFocus()
        QGraphicsView.leaveEvent(self, event)

    def mouseMoveEvent(self, event):
        """
        Must reimplement mouseMoveEvent of QGraphicsView to allow
        ScrollHandDrag due to the fact that events are intercepted
        breaks this feature.
        """
        if self._transformEnable == True:
            if self.dragMode() == self._yesDrag:
                # Add stuff to handle the pan event
                xf = event.pos().x()
                yf = event.pos().y()
                factor = self.transform().m11()
                transform = self.sceneRootItem.transform()

                transform.translate((xf - self._x0)/factor,
                                    (yf - self._y0)/factor)
                self.sceneRootItem.setTransform(transform)
                self._x0 = xf
                self._y0 = yf
            elif self._dollyZoomEnable == True:
                self.dollyZoom(event)
        # adding this allows events to be passed to items underneath
        QGraphicsView.mouseMoveEvent(self, event)
    # end def

    def mousePressEvent(self, event):
        """docstring for mousePressEvent"""
        if self._transformEnable == True and QGuiApplication.keyboardModifiers():
            which_buttons = event.buttons()
            if which_buttons in [self._button_pan, self._button_pan_alt]:
                self._panEnable()
                self._x0 = event.pos().x()
                self._y0 = event.pos().y()
            elif which_buttons == self._button_zoom:
                self._dollyZoomEnable = True
                self._last_scale_factor = 0
                # QMouseEvent.y() returns the position of the mouse cursor
                # relative to the widget
                self._y0 = event.posF().y()
            else:
                QGraphicsView.mousePressEvent(self, event)
        else:
            QGraphicsView.mousePressEvent(self, event)
    #end def

    def mouseReleaseEvent(self, event):
        """If panning, stop. If handles were pressed, release them."""
        if self._transformEnable == True:
            # QMouseEvent.button() returns the button that triggered the event
            which_button = event.button()
            if which_button in [self._button_pan, self._button_pan_alt]:
                self._panDisable()
            elif which_button == self._button_zoom:
                self._dollyZoomEnable = False
            else:
                return QGraphicsView.mouseReleaseEvent(self, event)
        # end if
        else:
            if len(self._pressList):  # Notify any pressed items to release
                # event_pos = event.position()
                for item in self._pressList:
                    #try:
                    # print "item release", item
                    item.customMouseRelease(event)
                    #except:
                    #    item.mouseReleaseEvent(event)
                #end for
                self._pressList = []
            # end if
            if self._selectionLock:
                self._selectionLock.processPendingToAddList()
            return QGraphicsView.mouseReleaseEvent(self, event)

    #end def

    def _panEnable(self):
        """Enable ScrollHandDrag Mode in QGraphicsView (displays a hand
        pointer)"""
        self.setDragMode(self._yesDrag)
    # end def

    def _panDisable(self):
        """Disable ScrollHandDrag Mode in QGraphicsView (displays a hand
        pointer)"""
        self.setDragMode(self._noDrag)
    # end def

    def fname(self):
        """docstring for fname"""
        pass

    def wheelEvent(self, event):
        self.safeScale(event.angleDelta().y())
    # end def

    def safeScale(self, delta):
        currentScaleLevel = self.transform().m11()
        scaleFactor = 1 + delta * \
           (self._scaleDownRate if delta < 0 else self._scaleUpRate) * \
           (app().prefs.zoomSpeed/100.)
        newScaleLevel = currentScaleLevel * scaleFactor
        newScaleLevel = util.clamp(currentScaleLevel * scaleFactor,\
                              self._scale_limit_min,\
                              self._scale_limit_max)
        scaleChange = newScaleLevel / currentScaleLevel
        self.scale(scaleChange, scaleChange)

        self.resetGL()
    # end def

    def zoomIn(self, fractionOfMax=0.5):
        currentScaleLevel = self.transform().m11()
        scaleChange = (fractionOfMax * self._scale_limit_max) / currentScaleLevel
        self.scale(scaleChange, scaleChange)
    # end def

    def zoomOut(self, fractionOfMin=1):
        currentScaleLevel = self.transform().m11()
        scaleChange = (fractionOfMin * self._scale_limit_min) / currentScaleLevel
        self.scale(scaleChange, scaleChange)
    # end def

    def dollyZoom(self, event):
        """docstring for dollyZoom"""
        # QMouseEvent.y() returns the position of the mouse cursor relative
        # to the widget
        yf = event.y()
        denom = abs(yf - self._y0)
        if denom > 0:
            scale_factor = (self.height() / 2) % denom
            if self._last_scale_factor != scale_factor:
                self._last_scale_factor = scale_factor
                # zoom in if mouse y position is getting bigger
                if yf - self._y0 > 0:
                    self.safeScale(yf - self._y0)
                # end else
                else:  # else id smaller zoom out
                    self.safeScale(yf - self._y0)
                # end else
        # end if
    # end def

    def resetScale(self):
        """
        reset the scale to 1
        """
        # use the transform value if you want to get how much the view
        # has been scaled
        self._scale_size = self.transform().m11()

        # self._scale_limit_min = 0.41*self._scale_size
        # make it so fitting in view is zoomed minimum
        # still gives you one zoom level out before violates limit
        self._scale_limit_min = self._scale_size*self._scaleFitFactor

        # use this if you want to reset the zoom in limit
        # self._scale_limit_max = 3.0*self._scale_size

        self._last_scale_factor = 0.0
    # end def

    def zoomToFit(self):
        # Auto zoom to center the scene
        thescene = self.sceneRootItem.scene()
        # order matters?
        self.sceneRootItem.resetTransform() # zero out translations
        self.resetTransform() # zero out scaling
        if self.toolbar:  # HACK: move toolbar so it doesn't affect sceneRect
            self.toolbar.setPos(0, 0)
        thescene.setSceneRect(thescene.itemsBoundingRect())
        scene_rect = thescene.sceneRect()
        if self.toolbar:  # HACK, pt2: move toolbar back
            self.toolbar.setPos(self.mapToScene(0, 0))
        self.fitInView(scene_rect, Qt.AspectRatioMode.KeepAspectRatio) # fit in view
        self.resetScale() # adjust scaling so that translation works
        # adjust scaling so that the items don't fill 100% of the view
        # this is good for selection
        self.scale(self._scaleFitFactor, self._scaleFitFactor)
        self._scale_size *= self._scaleFitFactor

        self.resetGL()
    # end def

    def paintEvent(self, event):
        if self.toolbar:
            self.toolbar.setPos(self.mapToScene(0, 0))
        QGraphicsView.paintEvent(self, event)

    def dragEnterEvent(self, event):
        event.ignore()
#end class
